import numpy as np

import _gpaw
from gpaw.utilities.blas import axpy
from gpaw import gpu


def multi_axpy_cpu(a, x, y):
    for ai, xi, yi in zip(a, x, y):
        axpy(ai, xi, yi)


def multi_axpy(a, x, y):
    assert type(x) is type(y)

    if isinstance(a, (float, complex)):
        axpy(a, x, y)
    else:
        if not isinstance(x, np.ndarray):
            if not isinstance(a, np.ndarray):
                a_gpu = a
            else:
                a_gpu = gpu.copy_to_device(a)
            _gpaw.multi_axpy_gpu(gpu.get_pointer(a_gpu),
                                 a.dtype,
                                 gpu.get_pointer(x),
                                 x.shape,
                                 gpu.get_pointer(y),
                                 y.shape,
                                 x.dtype)
        else:
            multi_axpy_cpu(a, x, y)



import numpy as np
import ase.units as units

from gpaw.utilities import unpack
from gpaw.wavefunctions.pw import PWWaveFunctions


def calculate_stress(calc):
    wfs = calc.wfs
    dens = calc.density
    ham = calc.hamiltonian

    if not isinstance(wfs, PWWaveFunctions):
        raise NotImplementedError('Calculation of stress tensor is only ' +
                                  'implemented for plane-wave mode.')
    if ham.xc.orbital_dependent and ham.xc.type != 'MGGA':
        raise NotImplementedError('Calculation of stress tensor is not ' +
                                  'implemented for orbital-dependent ' +
                                  'XC functionals such as ' + ham.xc.name)

    assert not ham.xc.name.startswith('GLLB')

    calc.timer.start('Stress tensor')

    s_vv = wfs.get_kinetic_stress().real
    nt_xg = dens.nt_xg
    if ham.xc_redistributor is not None:
        nt_xg = ham.xc_redistributor.distribute(nt_xg)
    s_vv += ham.xc.stress_tensor_contribution(nt_xg)

    pd = dens.pd3
    p_G = 4 * np.pi * dens.rhot_q
    G0 = 0 if pd.gd.comm.rank > 0 else 1
    p_G[G0:] /= pd.G2_qG[0][G0:]**2
    G_Gv = pd.get_reciprocal_vectors(add_q=False)
    for v1 in range(3):
        for v2 in range(3):
            s_vv[v1, v2] += pd.integrate(p_G, dens.rhot_q *
                                         G_Gv[:, v1] * G_Gv[:, v2])
    s_vv += dens.ghat.stress_tensor_contribution(ham.vHt_q, dens.Q_aL)

    s_vv -= np.eye(3) * ham.estress
    s_vv += ham.vbar.stress_tensor_contribution(dens.nt_Q)
    s_vv += dens.nct.stress_tensor_contribution(ham.vt_Q)
    if ham.xc.type == 'MGGA':
        nspin = ham.xc.dedtaut_sG.shape[0]
        vtau_sQ = dens.pd2.empty((nspin,), global_array=False)
        for s in range(nspin):
            vtau_sQ[s] = dens.pd2.fft(ham.xc.dedtaut_sG[s])
        s_vv += ham.xc.tauct.stress_tensor_contribution(vtau_sQ.mean(axis=0))

    s0 = 0.0
    s0_vv = 0.0
    for kpt in wfs.kpt_u:
        a_ani = {}
        for a, P_ni in kpt.P_ani.items():
            Pf_ni = P_ni * kpt.f_n[:, None]
            dH_ii = unpack(ham.dH_asp[a][kpt.s])
            dS_ii = ham.setups[a].dO_ii
            a_ni = (np.dot(Pf_ni, dH_ii) -
                    np.dot(Pf_ni * kpt.eps_n[:, None], dS_ii))
            s0 += np.vdot(P_ni, a_ni)
            a_ani[a] = 2 * a_ni.conj()
        s0_vv += wfs.pt.stress_tensor_contribution(kpt.psit_nG, a_ani,
                                                   q=kpt.q)
    s0_vv -= dens.gd.comm.sum(s0.real) * np.eye(3)
    s0_vv /= dens.gd.comm.size
    wfs.world.sum(s0_vv)
    s_vv += s0_vv

    s_vv += wfs.dedepsilon * np.eye(3)

    vol = calc.atoms.get_volume() / units.Bohr**3
    s_vv = 0.5 / vol * (s_vv + s_vv.T)

    # Symmetrize:
    sigma_vv = np.zeros((3, 3))
    cell_cv = wfs.gd.cell_cv
    for U_cc in wfs.kd.symmetry.op_scc:
        M_vv = np.dot(np.linalg.inv(cell_cv),
                      np.dot(U_cc, cell_cv)).T
        sigma_vv += np.dot(np.dot(M_vv.T, s_vv), M_vv)
    sigma_vv /= len(wfs.kd.symmetry.op_scc)

    # Make sure all agree on the result (redundant calculation on
    # different cores involving BLAS might give slightly different
    # results):
    wfs.world.broadcast(sigma_vv, 0)

    calc.log('Stress tensor:')
    for sigma_v in sigma_vv:
        calc.log('{:13.6f}{:13.6f}{:13.6f}'
                 .format(*(units.Ha / units.Bohr**3 * sigma_v)))

    calc.timer.stop('Stress tensor')

    return sigma_vv



import pytest
import numpy as np
from ase.build import bulk
from gpaw import GPAW


def run(xc, repeat=1):
    atoms = bulk('Ag') * repeat
    k = 4 // repeat
    calc = GPAW(mode='lcao',
                basis='sz(dzp)',
                h=0.3,
                setups={'Ag': '11'},
                nbands=6 * repeat**3,
                xc=xc,
                parallel={'domain': 1},
                kpts={'size': (k, k, k), 'gamma': False},
                txt='-')
    atoms.calc = calc
    atoms.get_potential_energy()
    x_i, y_i = calc.get_dos(npts=1001)
    x_i -= calc.get_fermi_level()
    y_i /= repeat**3
    return x_i, y_i


@pytest.mark.gllb
@pytest.mark.libxc
def test_metallic_GLLBSCM():
    # GLLBSC should behave like GLLBSCM for metals and
    # repeated cell should give exactly the same results
    x1_i, y1_i = run(xc='GLLBSCM', repeat=1)
    for x_i, y_i in [run(xc='GLLBSC', repeat=1),
                     run(xc='GLLBSCM', repeat=2)]:
        # Test that the DOSes are the same
        assert np.allclose(x1_i, x_i, rtol=0, atol=1e-8), \
            "DOS energies don't match, " \
            "error = {}".format(np.max(np.abs(x1_i - x_i)))
        assert np.allclose(y1_i, y_i, rtol=0, atol=1e-6), \
            "DOS values don't match, " \
            "error = {}".format(np.max(np.abs(y1_i - y_i)))



import numpy as np
from ase.units import Bohr, Hartree
from ase.parallel import paropen
from ase.data import vdw_radii

import _gpaw
from gpaw.io.fmf import FMF


class ExteriorElectronDensity:

    """Exterior electron density to describe MIES spectra.

    Simple approach to describe MIES spectra after
    Y. Harada et al., Chem. Rev. 97 (1997) 1897
    """

    def __init__(self, gd, atoms):
        """Find the grid points outside of the van der Waals radii
        of the atoms"""

        assert gd.orthogonal
        self.gd = gd

        n = len(atoms)
        atom_c = atoms.positions / Bohr
        vdWradius = np.empty((n))
        for a, atom in enumerate(atoms):
            vdWradius[a] = self.get_vdWradius(atom.number)

        # define the exterior region mask
        mask = gd.empty(dtype=int)
        _gpaw.eed_region(mask, atom_c, gd.beg_c, gd.end_c,
                         gd.h_cv.diagonal().copy(), vdWradius)
        self.mask = mask

    def get_weight(self, psit_G):
        """Get the weight of a wave function in the exterior region
        (outside of the van der Waals radius). The augmentation sphere
        is assumed to be smaller than the van der Waals radius and hence
        does not contribute."""

        # smooth part
        weigth = self.gd.integrate(np.where(self.mask == 1,
                                            (psit_G * psit_G.conj()).real,
                                            0.0))

        return weigth

    def get_vdWradius(self, Z):
        """Return van der Waals radius in Bohr"""
        r = vdw_radii[Z] / Bohr
        if np.isnan(r):
            msg = 'van der Waals radius for Z=' + str(Z) + ' not known!'
            raise RuntimeError(msg)
        else:
            return r

    def write_mies_weights(self, wfs):
        file = 'eed_mies.dat'
        with paropen(file, 'a') as out:
            fmf = FMF(['exterior electron density weights after',
                       'Y. Harada et al., Chem. Rev. 97 (1997) 1897'])
            print(fmf.header(), end=' ', file=out)
            print(fmf.data(['band index: n',
                            'k-point index: k',
                            'spin index: s',
                            'k-point weight: weight',
                            'energy: energy [eV]',
                            'occupation number: occ',
                            'relative EED weight: eed_weight']), end=' ',
                  file=out)

            print(
                '#; n   k s   weight      energy         occ  eed_weight',
                file=out)
            for kpt in wfs.kpt_u:
                for n in range(wfs.bd.nbands):
                    print('%4d %3d %1d %8.5f  %10.5f  %10.5f  %10.5f' %
                          (n, kpt.k, kpt.s, kpt.weight,
                           kpt.eps_n[n] * Hartree,
                           kpt.f_n[n],
                           self.get_weight(kpt.psit_nG[n])
                           ), file=out)
                    if hasattr(out, 'flush'):
                        out.flush()



import numpy as np
from scipy.linalg import eigh

from ase.units import Hartree
from ase.utils.timing import Timer

import gpaw.mpi as mpi
from gpaw.lrtddft.kssingle import KSSingles, KSSRestrictor
from gpaw.transformers import Transformer
from gpaw.utilities import pack
from gpaw.xc import XC

"""This module defines a Omega Matrix class."""


class OmegaMatrix:

    """
    Omega matrix in Casidas linear response formalism

    Parameters
      - calculator: the calculator object the ground state calculation
      - kss: the Kohn-Sham singles object
      - xc: the exchange correlation approx. to use
      - derivativeLevel: which level i of d^i Exc/dn^i to use
      - numscale: numeric epsilon for derivativeLevel=0,1
      - filehandle: the oject can be read from a filehandle
      - txt: output stream or file name
      - finegrid: level of fine grid to use. 0: nothing, 1 for poisson only,
        2 everything on the fine grid
    """

    def __init__(self,
                 calculator=None,
                 kss=None,
                 xc=None,
                 derivativeLevel=None,
                 numscale=0.001,
                 poisson=None,
                 filehandle=None,
                 log=None,
                 finegrid=2,
                 eh_comm=None):
        self.log = log

        if eh_comm is None:
            eh_comm = mpi.serial_comm

        self.eh_comm = eh_comm
        self.fullkss = kss

        if filehandle is not None:
            self.read(fh=filehandle)
            return None

        self.finegrid = finegrid

        if calculator is None:
            return

        self.paw = calculator
        wfs = self.paw.wfs

        # handle different grid possibilities
        self.restrict = None
        # self.poisson = PoissonSolver(nn=self.paw.hamiltonian.poisson.nn)
        if poisson is None:
            self.poisson = calculator.hamiltonian.poisson
        else:
            self.poisson = poisson
        if finegrid:
            self.poisson.set_grid_descriptor(self.paw.density.finegd)

            self.gd = self.paw.density.finegd
            if finegrid == 1:
                self.gd = wfs.gd
        else:
            self.poisson.set_grid_descriptor(wfs.gd)
            self.gd = wfs.gd
        self.restrict = Transformer(self.paw.density.finegd, wfs.gd,
                                    self.paw.density.stencil).apply

        if xc == 'RPA':
            xc = None  # enable RPA as keyword
        if xc is not None:
            if isinstance(xc, str):
                self.xc = XC(xc)
            else:
                self.xc = xc
            self.xc.initialize(self.paw.density, self.paw.hamiltonian, wfs)

            # check derivativeLevel
            if derivativeLevel is None:
                derivativeLevel = \
                    self.xc.get_functional().get_max_derivative_level()
            self.derivativeLevel = derivativeLevel
        else:
            self.xc = None

        self.numscale = numscale

        self.singletsinglet = False
        if kss.nvspins < 2 and kss.npspins < 2:
            # this will be a singlet to singlet calculation only
            self.singletsinglet = True

        nij = len(kss)
        self.Om = np.zeros((nij, nij))
        self.get_full()

    def get_full(self):
        """Evaluate full omega matrix"""
        self.paw.timer.start('Omega RPA')
        self.log()
        self.log('Linear response TDDFT calculation')
        self.log()
        self.get_rpa()
        self.paw.timer.stop()

        if self.xc is not None:
            self.paw.timer.start('Omega XC')
            self.get_xc()
            self.paw.timer.stop()

        self.eh_comm.sum(self.Om)
        self.full = self.Om

    def get_xc(self):
        """Add xc part of the coupling matrix"""

        # shorthands
        paw = self.paw
        wfs = paw.wfs
        gd = paw.density.finegd
        eh_comm = self.eh_comm

        fg = self.finegrid == 2
        kss = self.fullkss
        nij = len(kss)

        Om_xc = self.Om
        # initialize densities
        # nt_sg is the smooth density on the fine grid with spin index

        if kss.nvspins == 2:
            # spin polarised ground state calc.
            nt_sg = paw.density.nt_sg
        else:
            # spin unpolarised ground state calc.
            if kss.npspins == 2:
                # construct spin polarised densities
                nt_sg = np.array([.5 * paw.density.nt_sg[0],
                                  .5 * paw.density.nt_sg[0]])
            else:
                nt_sg = paw.density.nt_sg
        # check if D_sp have been changed before
        D_asp = {}
        for a, D_sp in self.paw.density.D_asp.items():
            if len(D_sp) != kss.npspins:
                if len(D_sp) == 1:
                    D_asp[a] = np.array([0.5 * D_sp[0], 0.5 * D_sp[0]])
                else:
                    D_asp[a] = np.array([D_sp[0] + D_sp[1]])
            else:
                D_asp[a] = D_sp.copy()

        # restrict the density if needed
        if fg:
            nt_s = nt_sg
        else:
            nt_s = self.gd.zeros(nt_sg.shape[0])
            for s in range(nt_sg.shape[0]):
                self.restrict(nt_sg[s], nt_s[s])
            gd = paw.density.gd

        # initialize vxc or fxc

        if self.derivativeLevel == 0:
            raise NotImplementedError
            if kss.npspins == 2:
                v_g = nt_sg[0].copy()
            else:
                v_g = nt_sg.copy()
        elif self.derivativeLevel == 1:
            pass
        elif self.derivativeLevel == 2:
            fxc_sg = np.zeros(nt_sg.shape)
            self.xc.calculate_fxc(gd, nt_sg, fxc_sg)
        else:
            raise ValueError('derivativeLevel can only be 0,1,2')

        ns = self.numscale
        xc = self.xc
        self.log('XC', nij, 'transitions')
        for ij in range(eh_comm.rank, nij, eh_comm.size):
            self.log('XC kss[' + '%d' % ij + ']')

            timer = Timer()
            timer.start('init')
            timer2 = Timer()

            if self.derivativeLevel >= 1:
                # vxc is available
                # We use the numerical two point formula for calculating
                # the integral over fxc*n_ij. The results are
                # vvt_s        smooth integral
                # nucleus.I_sp atom based correction matrices (pack2)
                #              stored on each nucleus
                timer2.start('init v grids')
                vp_s = np.zeros(nt_s.shape, nt_s.dtype.char)
                vm_s = np.zeros(nt_s.shape, nt_s.dtype.char)
                if kss.npspins == 2:  # spin polarised
                    nv_s = nt_s.copy()
                    nv_s[kss[ij].pspin] += ns * kss[ij].get(fg)
                    xc.calculate(gd, nv_s, vp_s)
                    nv_s = nt_s.copy()
                    nv_s[kss[ij].pspin] -= ns * kss[ij].get(fg)
                    xc.calculate(gd, nv_s, vm_s)
                else:  # spin unpolarised
                    nv = nt_s + ns * kss[ij].get(fg)
                    xc.calculate(gd, nv, vp_s)
                    nv = nt_s - ns * kss[ij].get(fg)
                    xc.calculate(gd, nv, vm_s)
                vvt_s = (0.5 / ns) * (vp_s - vm_s)
                timer2.stop()

                # initialize the correction matrices
                timer2.start('init v corrections')
                I_asp = {}
                for a, P_ni in wfs.kpt_u[kss[ij].spin].P_ani.items():
                    # create the modified density matrix
                    Pi_i = P_ni[kss[ij].i]
                    Pj_i = P_ni[kss[ij].j]
                    P_ii = np.outer(Pi_i, Pj_i)
                    # we need the symmetric form, hence we can pack
                    P_p = pack(P_ii)
                    D_sp = D_asp[a].copy()
                    D_sp[kss[ij].pspin] -= ns * P_p
                    setup = wfs.setups[a]
                    I_sp = np.zeros_like(D_sp)
                    self.xc.calculate_paw_correction(setup, D_sp, I_sp)
                    I_sp *= -1.0
                    D_sp = D_asp[a].copy()
                    D_sp[kss[ij].pspin] += ns * P_p
                    self.xc.calculate_paw_correction(setup, D_sp, I_sp)
                    I_sp /= 2.0 * ns
                    I_asp[a] = I_sp
                timer2.stop()

            timer.stop()
            t0 = timer.get_time('init')
            timer.start(ij)

            for kq in range(ij, nij):
                weight = self.weight_Kijkq(ij, kq)

                if self.derivativeLevel == 0:
                    # only Exc is available

                    if kss.npspins == 2:  # spin polarised
                        nv_g = nt_sg.copy()
                        nv_g[kss[ij].pspin] += kss[ij].get(fg)
                        nv_g[kss[kq].pspin] += kss[kq].get(fg)
                        Excpp = xc.get_energy_and_potential(
                            nv_g[0], v_g, nv_g[1], v_g)
                        nv_g = nt_sg.copy()
                        nv_g[kss[ij].pspin] += kss[ij].get(fg)
                        nv_g[kss[kq].pspin] -= kss[kq].get(fg)
                        Excpm = xc.get_energy_and_potential(
                            nv_g[0], v_g, nv_g[1], v_g)
                        nv_g = nt_sg.copy()
                        nv_g[kss[ij].pspin] -=\
                            kss[ij].get(fg)
                        nv_g[kss[kq].pspin] +=\
                            kss[kq].get(fg)
                        Excmp = xc.get_energy_and_potential(
                            nv_g[0], v_g, nv_g[1], v_g)
                        nv_g = nt_sg.copy()
                        nv_g[kss[ij].pspin] -= \
                            kss[ij].get(fg)
                        nv_g[kss[kq].pspin] -=\
                            kss[kq].get(fg)
                        Excpp = xc.get_energy_and_potential(
                            nv_g[0], v_g, nv_g[1], v_g)
                    else:  # spin unpolarised
                        nv_g = nt_sg + ns * kss[ij].get(fg)\
                            + ns * kss[kq].get(fg)
                        Excpp = xc.get_energy_and_potential(nv_g, v_g)
                        nv_g = nt_sg + ns * kss[ij].get(fg)\
                            - ns * kss[kq].get(fg)
                        Excpm = xc.get_energy_and_potential(nv_g, v_g)
                        nv_g = nt_sg - ns * kss[ij].get(fg)\
                            + ns * kss[kq].get(fg)
                        Excmp = xc.get_energy_and_potential(nv_g, v_g)
                        nv_g = nt_sg - ns * kss[ij].get(fg)\
                            - ns * kss[kq].get(fg)
                        Excmm = xc.get_energy_and_potential(nv_g, v_g)

                    Om_xc[ij, kq] += weight *\
                        0.25 * \
                        (Excpp - Excmp - Excpm + Excmm) / (ns * ns)

                elif self.derivativeLevel == 1:
                    # vxc is available

                    timer2.start('integrate')
                    Om_xc[ij, kq] += weight *\
                        self.gd.integrate(kss[kq].get(fg) *
                                          vvt_s[kss[kq].pspin])
                    timer2.stop()

                    timer2.start('integrate corrections')
                    Exc = 0.
                    for a, P_ni in wfs.kpt_u[kss[kq].spin].P_ani.items():
                        # create the modified density matrix
                        Pk_i = P_ni[kss[kq].i]
                        Pq_i = P_ni[kss[kq].j]
                        P_ii = np.outer(Pk_i, Pq_i)
                        # we need the symmetric form, hence we can pack
                        # use pack as I_sp used pack2
                        P_p = pack(P_ii)
                        Exc += np.dot(I_asp[a][kss[kq].pspin], P_p)
                    Om_xc[ij, kq] += weight * self.gd.comm.sum(Exc)
                    timer2.stop()

                elif self.derivativeLevel == 2:
                    # fxc is available
                    if kss.npspins == 2:  # spin polarised
                        Om_xc[ij, kq] += weight *\
                            gd.integrate(kss[ij].get(fg) *
                                         kss[kq].get(fg) *
                                         fxc_sg[kss[ij].pspin, kss[kq].pspin])
                    else:  # spin unpolarised
                        Om_xc[ij, kq] += weight *\
                            gd.integrate(kss[ij].get(fg) *
                                         kss[kq].get(fg) *
                                         fxc_sg)

                    # XXX still numeric derivatives for local terms
                    timer2.start('integrate corrections')
                    Exc = 0.
                    for a, P_ni in wfs.kpt_u[kss[kq].spin].P_ani.items():
                        # create the modified density matrix
                        Pk_i = P_ni[kss[kq].i]
                        Pq_i = P_ni[kss[kq].j]
                        P_ii = np.outer(Pk_i, Pq_i)
                        # we need the symmetric form, hence we can pack
                        # use pack as I_sp used pack2
                        P_p = pack(P_ii)
                        Exc += np.dot(I_asp[a][kss[kq].pspin], P_p)
                    Om_xc[ij, kq] += weight * self.gd.comm.sum(Exc)
                    timer2.stop()

                if ij != kq:
                    Om_xc[kq, ij] = Om_xc[ij, kq]

            timer.stop()
# timer2.write()
            if ij < (nij - 1):
                self.log('XC estimated time left',
                         self.time_left(timer, t0, ij, nij))

    def Coulomb_integral_kss(self, kss_ij, kss_kq, phit, rhot,
                             timer=None, yukawa=False):
        # smooth part
        if timer:
            timer.start('integrate')
        I = self.gd.integrate(rhot * phit)
        if timer:
            timer.stop()
            timer.start('integrate corrections 2')

        wfs = self.paw.wfs
        Pij_ani = wfs.kpt_u[kss_ij.spin].P_ani
        Pkq_ani = wfs.kpt_u[kss_kq.spin].P_ani

        # Add atomic corrections
        Ia = 0.0
        for a, Pij_ni in Pij_ani.items():
            Pi_i = Pij_ni[kss_ij.i]
            Pj_i = Pij_ni[kss_ij.j]
            Dij_ii = np.outer(Pi_i, Pj_i)
            Dij_p = pack(Dij_ii)
            Pk_i = Pkq_ani[a][kss_kq.i]
            Pq_i = Pkq_ani[a][kss_kq.j]
            Dkq_ii = np.outer(Pk_i, Pq_i)
            Dkq_p = pack(Dkq_ii)
            if yukawa and hasattr(self.xc, 'omega') and self.xc.omega > 0:
                C_pp = wfs.setups[a].calculate_yukawa_interaction(
                    self.xc.omega)
            else:
                C_pp = wfs.setups[a].M_pp
            #   ----
            # 2 >      P   P  C    P  P
            #   ----    ip  jr prst ks qt
            #   prst
            Ia += 2.0 * np.dot(Dkq_p, np.dot(C_pp, Dij_p))
        I += self.gd.comm.sum(Ia)
        if timer:
            timer.stop()

        return I

    def get_rpa(self):
        """calculate RPA part of the omega matrix"""

        # shorthands
        kss = self.fullkss
        finegrid = self.finegrid
        eh_comm = self.eh_comm

        # calculate omega matrix
        nij = len(kss)
        self.log('RPA', nij, 'transitions')

        Om = self.Om

        for ij in range(eh_comm.rank, nij, eh_comm.size):
            self.log('RPA kss[' + '%d' % ij + ']=', kss[ij])

            timer = Timer()
            timer.start('init')
            timer2 = Timer()

            # smooth density including compensation charges
            timer2.start('with_compensation_charges 0')
            rhot_p = kss[ij].with_compensation_charges(
                finegrid != 0)
            timer2.stop()

            # integrate with 1/|r_1-r_2|
            timer2.start('poisson')
            phit_p = np.zeros(rhot_p.shape, rhot_p.dtype.char)
            self.poisson.solve(phit_p, rhot_p, charge=None)
            timer2.stop()

            timer.stop()
            t0 = timer.get_time('init')
            timer.start(ij)

            if finegrid == 1:
                rhot = kss[ij].with_compensation_charges()
                phit = self.gd.zeros()
                self.restrict(phit_p, phit)
            else:
                phit = phit_p
                rhot = rhot_p

            for kq in range(ij, nij):
                if kq != ij:
                    # smooth density including compensation charges
                    timer2.start('kq with_compensation_charges')
                    rhot = kss[kq].with_compensation_charges(
                        finegrid == 2)
                    timer2.stop()

                pre = 2 * np.sqrt(kss[ij].get_energy() *
                                  kss[kq].get_energy() *
                                  kss[ij].get_weight() *
                                  kss[kq].get_weight())
                I = self.Coulomb_integral_kss(kss[ij], kss[kq],
                                              rhot, phit, timer2)
                Om[ij, kq] = pre * I

                if ij == kq:
                    Om[ij, kq] += kss[ij].get_energy() ** 2
                else:
                    Om[kq, ij] = Om[ij, kq]

            timer.stop()
# timer2.write()
            if ij < (nij - 1):
                t = timer.get_time(ij)  # time for nij-ij calculations
                t = .5 * t * \
                    (nij - ij)  # estimated time for n*(n+1)/2, n=nij-(ij+1)
                self.log('RPA estimated time left',
                         self.timestring(t0 * (nij - ij - 1) + t))

    def singlets_triplets(self):
        """Split yourself into singlet and triplet transitions"""

        assert self.fullkss.npspins == 2
        assert self.fullkss.nvspins == 1

        # strip kss from down spins
        skss = KSSingles()
        skss.dtype = self.fullkss.dtype
        tkss = KSSingles()
        tkss.dtype = self.fullkss.dtype
        map = []
        for ij, ks in enumerate(self.fullkss):
            if ks.pspin == ks.spin:
                skss.append((ks + ks) / np.sqrt(2))
                tkss.append((ks - ks) / np.sqrt(2))
                map.append(ij)
        skss.istart = tkss.istart = self.fullkss.restrict['istart']
        skss.jend = tkss.jend = self.fullkss.restrict['jend']
        nkss = len(skss)

        # define the singlet and the triplet omega-matrices
        sOm = OmegaMatrix(kss=skss, log=self.log)
        sOm.full = np.empty((nkss, nkss))
        tOm = OmegaMatrix(kss=tkss, log=self.log)
        tOm.full = np.empty((nkss, nkss))
        for ij in range(nkss):
            for kl in range(nkss):
                sOm.full[ij, kl] = (self.full[map[ij], map[kl]] +
                                    self.full[map[ij], nkss + map[kl]])
                tOm.full[ij, kl] = (self.full[map[ij], map[kl]] -
                                    self.full[map[ij], nkss + map[kl]])
        return sOm, tOm

    def timestring(self, t):
        ti = int(t + 0.5)
        td = ti // 86400
        st = ''
        if td > 0:
            st += '%d' % td + 'd'
            ti -= td * 86400
        th = ti // 3600
        if th > 0:
            st += '%d' % th + 'h'
            ti -= th * 3600
        tm = ti // 60
        if tm > 0:
            st += '%d' % tm + 'm'
            ti -= tm * 60
        st += '%d' % ti + 's'
        return st

    def time_left(self, timer, t0, ij, nij):
        t = timer.get_time(ij)  # time for nij-ij calculations
        t = .5 * t * (nij - ij)  # estimated time for n*(n+1)/2, n=nij-(ij+1)
        return self.timestring(t0 * (nij - ij - 1) + t)

    def get_map(self, restrict={}):
        """Return the reduction map for the given requirements

        Returns
        -------
        map - list of original indices
        kss - reduced KSSingles object
        """
        rst = KSSRestrictor(restrict)
        if rst >= self.fullkss.restrict:
            return None, self.fullkss

        self.log('# diagonalize: %d transitions original'
                 % len(self.fullkss))

        map = []
        kss = KSSingles()
        kss.dtype = self.fullkss.dtype
        energy_range = rst['energy_range']
        emin, emax = rst.emin_emax()
        istart = rst['istart']
        jend = rst['jend']
        eps = rst['eps']
        for ij, k in zip(range(len(self.fullkss)), self.fullkss):
            if energy_range is None:
                if k.i >= istart and k.j <= jend and k.fij >= eps:
                    kss.append(k)
                    map.append(ij)
            else:
                if k.energy >= emin and k.energy < emax and k.fij >= eps:
                    kss.append(k)
                    map.append(ij)
        kss.update()
        self.log('# diagonalize: %d transitions now' % len(kss))

        return map, kss

    def diagonalize(self, restrict={}):
        """Evaluate Eigenvectors and Eigenvalues:"""
        map, kss = self.get_map(restrict)

        nij = len(kss)
        if map is None:
            evec = self.full.copy()
        else:
            evec = np.zeros((nij, nij))
            for ij in range(nij):
                for kq in range(nij):
                    evec[ij, kq] = self.full[map[ij], map[kq]]
        assert len(evec) > 0

        self.eigenvalues, v = eigh(evec)
        self.eigenvectors = v.T
        self.kss = kss

    @property
    def kss(self):
        if hasattr(self, '_kss'):
            return self._kss
        return self.fullkss

    @kss.setter
    def kss(self, kss):
        """Set current (restricted) KSSingles object"""
        self._kss = kss

    def read(self, filename=None, fh=None):
        """Read myself from a file"""
        if fh is None:
            f = open(filename, 'r')
        else:
            f = fh

        f.readline()
        nij = int(f.readline())
        full = np.zeros((nij, nij))
        for ij in range(nij):
            l = [float(x) for x in f.readline().split()]
            full[ij, ij:] = l
            full[ij:, ij] = l
        self.full = full

        if fh is None:
            f.close()

    def write(self, filename=None, fh=None):
        """Write current state to a file."""
        try:
            rank = self.paw.wfs.world.rank
        except AttributeError:
            rank = mpi.world.rank
        if rank == 0:
            if fh is None:
                f = open(filename, 'w')
            else:
                f = fh

            f.write('# OmegaMatrix\n')
            nij = len(self.fullkss)
            f.write('%d\n' % nij)
            for ij in range(nij):
                for kq in range(ij, nij):
                    f.write(' %g' % self.full[ij, kq])
                f.write('\n')

            if fh is None:
                f.close()

    def weight_Kijkq(self, ij, kq):
        """weight for the coupling matrix terms"""
        kss = self.fullkss
        return 2. * np.sqrt(kss[ij].get_energy() *
                            kss[kq].get_energy() *
                            kss[ij].get_weight() *
                            kss[kq].get_weight())

    def __str__(self):
        str = '<OmegaMatrix> '
        if hasattr(self, 'eigenvalues'):
            str += 'dimension ' + ('%d' % len(self.eigenvalues))
            str += '\neigenvalues: '
            for ev in self.eigenvalues:
                str += ' ' + ('%f' % (np.sqrt(ev) * Hartree))
        return str



from typing import Any, Dict, Optional, List, Sequence, Union

import numpy as np
from ase.units import Bohr
from ase.utils.timing import Timer
from gpaw.poisson import _PoissonSolver, create_poisson_solver
from gpaw.utilities.gauss import Gaussian
from gpaw.typing import Array1D
from gpaw.utilities.timing import nulltimer, NullTimer

from ase.utils.timing import timer


MomentCorrectionsType = Union[int, List[Dict[str, Any]]]


class MomentCorrection:

    def __init__(self,
                 center: Optional[Union[Sequence, Array1D]],
                 moms: Union[int, Sequence[int]]):
        if center is not None:
            center = np.asarray(center) / Bohr
        self.center = center
        self.moms = np.asarray(moms)

    def todict(self) -> Dict[str, Any]:
        """ return dictionary description, converting the moment correction
        from units of Bohr to Ångström """

        center = self.center
        if center is not None:
            center = center * Bohr

        dict_out = dict(moms=self.moms, center=center)

        return dict_out

    def __str__(self) -> str:
        if self.center is None:
            center = 'center'
        else:
            center = ', '.join([f'{x:.2f}' for x in self.center * Bohr])

        if np.allclose(np.diff(self.moms), 1):
            # Increasing sequence
            moms = f'range({self.moms[0]}, {self.moms[-1]+1})'
        else:
            # List of integers
            _moms = ', '.join([f'{m:.0f}' for m in self.moms])
            moms = f'({_moms})'
        return f'[{center}] {moms}'

    def __repr__(self) -> str:
        center = self.center
        if center is not None:
            center *= Bohr
        return f'{repr(self.moms)} @ {repr(center)}'


class MomentCorrectionPoissonSolver(_PoissonSolver):
    """Wrapper for the poisson solver that includes moment corrections

    Parameters
    ----------
    poissonsolver
        underlying poisson solver
    moment_corrections
        list of moment corrections, expressed as dictionaries
        `{'moms': ..., 'center': ...}` that specify the multipole moments
        and their centres.

        moment_corrections = [{'moms': moms_list1, 'center': center1},
                              {'moms': moms_list2, 'center': center2},
                              ...]

        Here moms_listX is list of integers of multipole moments to be
        corrected at centerX. For example moms_list=range(4) corresponds to
        s, p_x, p_y and p_z type multipoles.

        Optionally setting moment_corrections to an integer m is equivalent to
        including multipole moments corresponding to range(m) at the center of
        the cell


    timer
        timer

    """

    def __init__(self,
                 poissonsolver: Union[_PoissonSolver, Dict[str, Any]],
                 moment_corrections: Optional[MomentCorrectionsType],
                 timer: Union[NullTimer, Timer] = nulltimer):

        self._initialized = False
        self.poissonsolver = create_poisson_solver(poissonsolver)
        self.timer = timer

        if moment_corrections is None:
            self.moment_corrections = []
        elif isinstance(moment_corrections, int):
            moms = range(moment_corrections)
            center = None
            self.moment_corrections = [MomentCorrection(moms=moms,
                                                        center=center)]
        elif isinstance(moment_corrections, list):
            assert all(['moms' in mom and 'center' in mom
                        for mom in moment_corrections]), \
                   (f'{self.__class__.__name__}: each element in '
                    'moment_correction must be a dictionary '
                    'with the keys "moms" and "center"')
            self.moment_corrections = [MomentCorrection(**mom)
                                       for mom in moment_corrections]
        else:
            raise ValueError(f'{self.__class__.__name__}: moment_correction '
                             'must be a list of dictionaries')

    def todict(self):
        mom_corr = [mom.todict() for mom in self.moment_corrections]
        d = {'name': 'MomentCorrectionPoissonSolver',
             'poissonsolver': self.poissonsolver.todict(),
             'moment_corrections': mom_corr}

        return d

    def set_grid_descriptor(self, gd):
        self.poissonsolver.set_grid_descriptor(gd)
        self.gd = gd

    def get_description(self) -> str:
        description = self.poissonsolver.get_description()
        n = len(self.moment_corrections)

        lines = [description]
        lines.extend([f'    {n} moment corrections:'])
        lines.extend([f'      {str(mom)}'
                     for mom in self.moment_corrections])

        return '\n'.join(lines)

    @timer('Poisson initialize')
    def _init(self):
        if self._initialized:
            return
        self.poissonsolver._init()

        if not self.gd.orthogonal or self.gd.pbc_c.any():
            raise NotImplementedError('Only orthogonal unit cells '
                                      'and non-periodic boundary '
                                      'conditions are tested')
        self.load_moment_corrections_gauss()

        self._initialized = True

    @timer('Load moment corrections')
    def load_moment_corrections_gauss(self):
        self.gauss_i = []
        self.mom_ij = []
        self.mask_ig = []

        if len(self.moment_corrections) == 0:
            return

        mask_ir = []
        r_ir = []

        for rmom in self.moment_corrections:
            center = rmom.center
            mom_j = rmom.moms
            gauss = Gaussian(self.gd, center=center)
            self.gauss_i.append(gauss)
            r_ir.append(gauss.r.ravel())
            mask_ir.append(self.gd.zeros(dtype=int).ravel())
            self.mom_ij.append(mom_j)

        r_ir = np.array(r_ir)
        mask_ir = np.array(mask_ir)

        Ni = r_ir.shape[0]
        Nr = r_ir.shape[1]

        for r in range(Nr):
            i = np.argmin(r_ir[:, r])
            mask_ir[i, r] = 1

        for i in range(Ni):
            mask_r = mask_ir[i]
            mask_g = mask_r.reshape(self.gd.n_c)
            self.mask_ig.append(mask_g)

    def solve(self, phi, rho, **kwargs):
        self._init()
        return self._solve(phi, rho, **kwargs)

    @timer('Solve')
    def _solve(self, phi, rho, **kwargs):
        timer = kwargs.get('timer', self.timer)

        if len(self.moment_corrections) > 0:
            assert not self.gd.pbc_c.any()

            timer.start('Multipole moment corrections')

            rho_neutral = rho * 0.0
            phi_cor_g = self.gd.zeros()
            for gauss, mask_g, mom_j in zip(self.gauss_i, self.mask_ig,
                                            self.mom_ij):
                rho_masked = rho * mask_g
                for mom in mom_j:
                    phi_cor_g += gauss.remove_moment(rho_masked, mom)
                rho_neutral += rho_masked

            # Remove multipoles for better initial guess
            phi -= phi_cor_g

            timer.stop('Multipole moment corrections')

            timer.start('Solve neutral')
            niter = self.poissonsolver.solve(phi, rho_neutral, **kwargs)
            timer.stop('Solve neutral')

            timer.start('Multipole moment corrections')
            # correct error introduced by removing multipoles
            phi += phi_cor_g
            timer.stop('Multipole moment corrections')

            return niter
        else:
            return self.poissonsolver.solve(phi, rho, **kwargs)

    def estimate_memory(self, mem):
        self.poissonsolver.estimate_memory(mem)
        gdbytes = self.gd.bytecount()
        if self.moment_corrections is not None:
            mem.subnode('moment_corrections masks',
                        len(self.moment_corrections) * gdbytes)

    def __repr__(self):
        if len(self.moment_corrections) == 0:
            corrections_str = 'no corrections'
        elif len(self.moment_corrections) < 2:
            corrections_str = f'{repr(self.moment_corrections[0])}'
        else:
            corrections_str = f'{len(self.moment_corrections)} corrections'

        representation = f'MomentCorrectionPoissonSolver ({corrections_str})'
        return representation



import numpy as np

from gpaw.gaunt import gaunt
from gpaw.sphere.rshe import RealSphericalHarmonicsExpansion
from gpaw.sphere.integrate import radial_truncation_function


def calculate_site_matrix_element_correction(
        pawdata, rshe: RealSphericalHarmonicsExpansion,
        rcut_p, drcut, lambd_p):
    r"""Calculate the PAW correction to a site matrix element.

    For the site matrix element

    f^ap_nn' = <ψ_n|Θ(r∊Ω_ap)f(r)|ψ_n'>

    the PAW correction for each pair of partial waves i and i' is given by
               l
           __  __
     ap    \   \   m,mi,mi' /                        a    a     ̰ a   ̰ a
    F    = /   /  g         | r^2 dr θ(r<rc) f (r) [φ(r) φ(r) - φ(r) φ(r)]
     ii'   ‾‾  ‾‾  l,li,li' /         p       lm     i    i'     i    i'
           l  m=-l

    where g refer to the Gaunt coefficients and f_lm(r) are the real
    spherical harmonics expansion coefficients of the function f(r).

    Here, we evaluate the correction F_ii'^ap for various smooth truncation
    functions θ_p(r<rc), parametrized by rc, Δrc and λ.
    """
    rgd = rshe.rgd
    assert rgd is pawdata.xc_correction.rgd
    ni = pawdata.ni  # Number of partial waves
    l_j = pawdata.l_j  # l-index for each radial function index j
    lmax = max(l_j)
    G_LLL = gaunt(lmax)
    assert max(rshe.l_M) <= lmax * 2
    # (Real) radial functions for the partial waves
    phi_jg = pawdata.data.phi_jg
    phit_jg = pawdata.data.phit_jg
    # Truncate the radial functions to span only the radial grid coordinates
    # which need correction
    assert np.allclose(rgd.r_g - pawdata.rgd.r_g[:rgd.N], 0.)
    phi_jg = np.array(phi_jg)[:, :rgd.N]
    phit_jg = np.array(phit_jg)[:, :rgd.N]

    # Calculate smooth truncation functions and allocate array
    Np = len(rcut_p)
    assert len(lambd_p) == Np
    theta_pg = [radial_truncation_function(rgd.r_g, rcut, drcut, lambd)
                for rcut, lambd in zip(rcut_p, lambd_p)]
    F_pii = np.zeros((Np, ni, ni), dtype=float)

    # Loop of radial function indices for partial waves i and i'
    i1_counter = 0
    for j1, l1 in enumerate(l_j):
        i2_counter = 0
        for j2, l2 in enumerate(l_j):
            # Calculate the radial partial wave correction
            dn_g = phi_jg[j1] * phi_jg[j2] - phit_jg[j1] * phit_jg[j2]

            # Generate m-indices for each radial function
            for m1 in range(2 * l1 + 1):
                for m2 in range(2 * l2 + 1):
                    # Set up the i=(l,m) index for each partial wave
                    i1 = i1_counter + m1
                    i2 = i2_counter + m2

                    # Loop through the real spherical harmonics of the local
                    # function f(r)
                    for L, f_g in zip(rshe.L_M, rshe.f_gM.T):
                        # Angular integral
                        gaunt_coeff = G_LLL[l1**2 + m1, l2**2 + m2, L]
                        if gaunt_coeff == 0:
                            continue
                        # Radial integral
                        for p, theta_g in enumerate(theta_pg):
                            F_pii[p, i1, i2] += \
                                gaunt_coeff * rgd.integrate_trapz(
                                theta_g * f_g * dn_g)

            # Add to i and i' counters
            i2_counter += 2 * l2 + 1
        i1_counter += 2 * l1 + 1
    return F_pii



import numpy as np
import pytest
from ase.build import bulk

from gpaw import GPAW, PW, Mixer
from gpaw.mpi import world


@pytest.mark.later
@pytest.mark.stress
def test_pw_si_stress(in_tmp_dir, gpaw_new):
    xc = 'PBE'
    si = bulk('Si')
    si.calc = GPAW(mode=PW(200),
                   mixer=Mixer(0.7, 5, 50.0),
                   xc=xc,
                   kpts=(1, 1, 2),  # Run (1, 1, 2) to avoid gamma pt code
                   convergence={'energy': 1e-8},
                   parallel={'domain': min(2, world.size)},
                   txt='si_stress.txt')

    si.set_cell(np.dot(si.cell,
                       [[1.02, 0, 0.03],
                        [0, 0.99, -0.02],
                        [0.2, -0.01, 1.03]]),
                scale_atoms=True)

    si.get_potential_energy()

    if not gpaw_new:
        # Trigger nasty bug (fixed in !486):
        si.calc.wfs.pt.blocksize = si.calc.wfs.pd.maxmyng - 1

    s_analytical = si.get_stress()
    s_ref = [-0.16569446, -0.07630128, -0.1266625,
             -0.06144752, -0.02055657, 0.04574812]
    # si.calc.calculate_numerical_stress(si, 1e-5)
    print(s_analytical)
    s_err = s_analytical - s_ref
    assert np.all(abs(s_err) < 1e-4)



import numpy as np

from scipy.optimize import minimize

from gpaw.spline import Spline
from gpaw.lfc import LocalizedFunctionsCollection
from gpaw.sphere.lebedev import weight_n


def integrate_lebedev(f_nx):
    """Integrate the function f(r) on the angular Lebedev quadrature.

    Here, n is the quadrature index for the angular dependence of the function
    defined on a spherical grid, while x are some arbitrary extra dimensions.
    """
    return 4. * np.pi * np.tensordot(weight_n, f_nx, axes=([0], [0]))


def integrate_radial_grid(f_xg, r_g, rcut=None):
    """Integrate the function f(r) on the radial grid.

    Computes the integral

    /
    | r^2 dr f(r)
    /

    for the range of values r on the grid r_g (up to rcut, if specified).
    """
    if rcut is not None:
        f_xg, r_g = truncate_radial_grid(f_xg, r_g, rcut)

    # Perform actual integration using the radial trapezoidal rule
    f_x = radial_trapz(f_xg, r_g)

    return f_x


def truncate_radial_grid(f_xg, r_g, rcut):
    """Truncate the radial grid representation of a function f(r) at r=rcut.

    If rcut is not part of the original grid, it will be added as a grid point,
    with f(rcut) determined by linear interpolation."""
    assert rcut > 0.
    assert np.any(r_g >= rcut)
    if rcut not in r_g:
        f_gx = np.moveaxis(f_xg, -1, 0)
        # Find the two points closest to rcut and interpolate between them
        # to get the value at rcut
        g1, g2 = find_two_closest_grid_points(r_g, rcut)
        r1 = r_g[g1]
        r2 = r_g[g2]
        lambd = (rcut - r1) / (r2 - r1)
        f_interpolated_x = (1 - lambd) * f_gx[g1] + lambd * f_gx[g2]
        # Add rcut as a grid point
        r_g = np.append(r_g, np.array([rcut]))
        f_gx = np.append(f_gx, np.array([f_interpolated_x]), axis=0)
        f_xg = np.moveaxis(f_gx, 0, -1)
    # Pick out the grid points inside rcut
    mask_g = r_g <= rcut
    r_g = r_g[mask_g]
    f_xg = f_xg[..., mask_g]

    return f_xg, r_g


def find_two_closest_grid_points(r_g, rcut):
    """Find the two closest grid point to a specified rcut."""
    # Find the two smallest absolute differences
    abs_diff_g = abs(r_g - rcut)
    ad1, ad2 = np.partition(abs_diff_g, 1)[:2]

    # Identify the corresponding indices
    g1 = np.where(abs_diff_g == ad1)[0][0]
    g2 = np.where(abs_diff_g == ad2)[0][0]

    return g1, g2


def radial_trapz(f_xg, r_g):
    r"""Integrate the function f(r) using the radial trapezoidal rule.

    Linearly interpolating,

                    r - r0
    f(r) ≃ f(r0) + ‾‾‾‾‾‾‾ (f(r1) - f(r0))      for r0 <= r <= r1
                   r1 - r0

    the integral

    /
    | r^2 dr f(r)
    /

    can be constructed in a piecewise manner from each discretized interval
    r_(n-1) <= r <= r_n, using:

    r1
    /               1
    | r^2 dr f(r) ≃ ‾ (r1^3 f(r1) - r0^3 f(r0))
    /               4
    r0                r1^3 - r0^3
                    + ‾‾‾‾‾‾‾‾‾‾‾ (r1 f(r0) - r0 f(r1))
                      12(r1 - r0)
    """
    assert np.all(r_g >= 0.)
    assert f_xg.shape[-1] == len(r_g)

    # Start and end of each discretized interval
    r0_g = r_g[:-1]
    r1_g = r_g[1:]
    f0_xg = f_xg[..., :-1]
    f1_xg = f_xg[..., 1:]
    assert np.all(r1_g - r0_g > 0.), \
        'Please give the radial grid in ascending order'

    # Linearly interpolate f(r) between r0 and r1 and integrate r^2 f(r)
    # in this area
    integrand_xg = (r1_g**3. * f1_xg - r0_g**3. * f0_xg) / 4.
    integrand_xg += (r1_g**3. - r0_g**3.) * (r1_g * f0_xg - r0_g * f1_xg)\
        / (12. * (r1_g - r0_g))

    # Sum over the discretized integration intervals
    return np.sum(integrand_xg, axis=-1)


def radial_truncation_function(r_g, rcut, drcut=None, lambd=None):
    r"""Generate smooth radial truncation function θ(r<rc).

    The function is generated to interpolate smoothly between the values

           ( 1    for r <= rc - Δrc/2
    θ(r) = < λ    for r = rc
           ( 0    for r >= rc + Δrc/2

    In the interpolation region, rc - Δrc/2 < r < rc + Δrc/2, the nonanalytic
    smooth function

           ( exp(-1/x)  for x > 0
    f(x) = <
           ( 0          for x <= 0

    is used to define θ(r), in order for all derivatives to be continous:

                        f(1/2-[r-rc]/Δrc)
    θ(r) = ‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾
           f(1/2-[r-rc]/Δrc) + (1-λ)f(1/2+[r-rc]/Δrc)/λ

    Unless given as an input, we choose 0 < λ < 1 to conserve the spherical
    integration volume, 4π rc^3/3.
    """
    assert np.all(r_g >= 0.)
    if drcut is None:
        # As a default, define Δrc to match twice the grid sampling around the
        # cutoff
        g1, g2 = find_two_closest_grid_points(r_g, rcut)
        drcut = 2 * abs(r_g[g2] - r_g[g1])
    assert rcut > 0. and drcut > 0. and rcut - drcut / 2. >= 0.
    if lambd is None:
        lambd = find_volume_conserving_lambd(rcut, drcut, r_g)
    assert 0. < lambd and lambd < 1.
    assert np.any(r_g >= rcut + drcut / 2.)

    def f(x):
        out = np.zeros_like(x)
        out[x > 0] = np.exp(-1 / x[x > 0])
        return out

    # Create array of ones inside rc + Δrc/2
    theta_g = np.ones_like(r_g)
    theta_g[r_g >= rcut + drcut / 2.] = 0.

    # Add smooth truncation
    gmask = np.logical_and(rcut - drcut / 2. < r_g, r_g < rcut + drcut / 2.)
    tr_g = r_g[gmask]
    theta_g[gmask] = f(1 / 2. - (tr_g - rcut) / drcut) \
        / (f(1 / 2. - (tr_g - rcut) / drcut)
           + (1 - lambd) * f(1 / 2. + (tr_g - rcut) / drcut) / lambd)

    return theta_g


def find_volume_conserving_lambd(rcut, drcut, r_g=None):
    r"""Determine the scaling factor λ to conserve the spherical volume.

    For a given rc and drc, λ is determined to make θ(r) numerically satisfy:
       ∞
       /
    4π | r^2 dr θ(r) = 4π rc^3/3
       /
       0
    """
    if r_g is None:
        r_g = _uniform_radial_grid(rcut, drcut)
    ref = 4 * np.pi * rcut**3. / 3.

    def integration_volume_error(lambd):
        theta_g = radial_truncation_function(r_g, rcut, drcut, lambd)
        vol = 4 * np.pi * radial_trapz(theta_g, r_g)
        return (vol - ref)**2.

    opt_result = minimize(integration_volume_error,
                          1 / 2.,  # start guess
                          bounds=[(1e-8, 1 - 1e-8)],
                          method='L-BFGS-B',
                          options={'ftol': 1e-12})
    if opt_result.success:
        lambd = opt_result.x[0]
    else:
        raise Exception('Could not find an appropriate truncation scaling λ',
                        opt_result.message)

    return lambd


def periodic_truncation_function(gd, spos_c, rcut, drcut=None, lambd=None):
    r"""Generate periodic images of the spherical truncation function θ.

    The smooth radial truncation function θ(r<rc) is used to define a
    smoothly truncated sphere of radius rcut, centered at the scaled
    position spos_c. The sphere is periodically repeated on the real-space grid
    described by gd.
    """
    # Generate a spherical truncation function collection with a single
    # truncation function
    stfc = spherical_truncation_function_collection(
        gd, spos_ac=[spos_c],
        rcut_aj=[[rcut]], drcut=drcut, lambd_aj=[[lambd]])

    # Evaluate the spherical truncation function (with its periodic images) on
    # the real-space grid
    theta_R = gd.zeros(dtype=float)
    stfc.add(theta_R)

    return theta_R


def spherical_truncation_function_collection(gd, spos_ac, rcut_aj,
                                             drcut=None, lambd_aj=None,
                                             kd=None, dtype=float):
    """Generate a collection of spherical truncation functions θ(|r-r_a|<rc_aj)

    Generates a LocalizedFunctionsCollection with radial truncation functions
    θ(r<rc_aj), centered at the scaled positions spos_ac.

    See radial_truncation_function() for the functional form of θ(r<rc).
    """
    if drcut is None:
        drcut = default_spherical_drcut(gd)
    if lambd_aj is None:
        # Match the nested list structure of rcut_aj and determine the actual
        # lambda values later
        lambd_aj = [[None] * len(rcut_j) for rcut_j in rcut_aj]

    # Generate splines for each atomic site and radial truncation function
    spline_aj = []
    for spos_c, rcut_j, lambd_j in zip(spos_ac, rcut_aj, lambd_aj):
        spline_j = []
        for rcut, lambd in zip(rcut_j, lambd_j):
            spline_j.append(radial_truncation_function_spline(
                rcut, drcut, lambd))
        spline_aj.append(spline_j)

    # Generate the spherical truncation function collection (stfc)
    stfc = LocalizedFunctionsCollection(gd, spline_aj, kd=kd, dtype=dtype)
    stfc.set_positions(spos_ac)

    return stfc


def radial_truncation_function_spline(rcut, drcut, lambd=None):
    """Generate spline representation of the radial truncation function θ(r<rc)
    """
    if lambd is None:
        lambd = find_volume_conserving_lambd(rcut, drcut)

    # Lay out truncation function on a radial grid and generate spline
    r_g = _uniform_radial_grid(rcut, drcut)
    theta_g = radial_truncation_function(r_g, rcut, drcut, lambd)
    spline = Spline(l=0, rmax=max(r_g),
                    # The input f_g is the expansion coefficient of the
                    # requested spherical harmonic for the function in
                    # question. For l=0, Y = 1/sqrt(4π):
                    f_g=np.sqrt(4 * np.pi) * theta_g)
    return spline


def default_spherical_drcut(gd):
    """Define default width for the spherical truncation function."""
    # Find the side-length corresponding to a cubic grid volume element
    length = gd.dv**(1. / 3.)
    # Use this side-length as the default
    return length


def _uniform_radial_grid(rcut, drcut):
    return np.linspace(0., rcut + 2 * drcut, 251)


if __name__ == '__main__':
    import matplotlib.pyplot as plt

    r_g = np.linspace(0., 4., 200)

    plt.subplot(1, 2, 1)
    plt.plot(r_g, radial_truncation_function(r_g, 1.0, 1.0))
    plt.plot(r_g, radial_truncation_function(r_g, 2.0, 1.0))
    plt.plot(r_g, radial_truncation_function(r_g, 3.0, 1.0))
    plt.subplot(1, 2, 2)
    plt.plot(r_g, radial_truncation_function(r_g, 1.0, 2.0))
    plt.plot(r_g, radial_truncation_function(r_g, 2.0, 2.0))
    plt.plot(r_g, radial_truncation_function(r_g, 3.0, 2.0))
    plt.show()



import pytest
from gpaw import GPAW
from gpaw.response.paw import calculate_pair_density_correction
import numpy as np


@pytest.mark.response
def test_two_phi_integrals(gpw_files):
    calc = GPAW(gpw_files['bn_pw'])

    setup = calc.wfs.setups[0]
    k_Gv = np.array([[0.0, 0.0, 0.0]])
    dO_aii = calculate_pair_density_correction(k_Gv, pawdata=setup)
    assert dO_aii[0] == pytest.approx(setup.dO_ii, 1e-8, 1e-7)



from math import pi
import numpy as np

from gpaw.density import Density
from gpaw.pw.descriptor import PWDescriptor, PWMapping
from gpaw.pw.lfc import PWLFC


class PseudoCoreKineticEnergyDensityLFC(PWLFC):
    def add(self, tauct_R):
        tauct_R += self.pd.ifft(1.0 / self.pd.gd.dv *
                                self.expand().sum(1).view(complex))

    def derivative(self, dedtaut_R, dF_aiv):
        PWLFC.derivative(self, self.pd.fft(dedtaut_R), dF_aiv)


class ReciprocalSpaceDensity(Density):
    def __init__(self, ecut,
                 gd, finegd, nspins, collinear, charge, redistributor,
                 background_charge=None):
        Density.__init__(self, gd, finegd, nspins, collinear, charge,
                         redistributor=redistributor,
                         background_charge=background_charge)
        ecut0 = 0.5 * pi**2 / (gd.h_cv**2).sum(1).max()
        ecut = min(ecut, ecut0)
        self.pd2 = PWDescriptor(ecut, gd)
        self.pd3 = PWDescriptor(4 * ecut, finegd)

        self.map23 = PWMapping(self.pd2, self.pd3)

        self.nct_q = None
        self.nt_Q = None
        self.rhot_q = None

    def initialize(self, setups, timer, magmom_av, hund):
        Density.initialize(self, setups, timer, magmom_av, hund)

        spline_aj = []
        for setup in setups:
            if setup.nct is None:
                spline_aj.append([])
            else:
                spline_aj.append([setup.nct])
        self.nct = PWLFC(spline_aj, self.pd2)

        self.ghat = PWLFC([setup.ghat_l for setup in setups], self.pd3,
                          )  # blocksize=256, comm=self.xc_redistributor.comm)

    def set_positions(self, spos_ac, atom_partition):
        Density.set_positions(self, spos_ac, atom_partition)
        self.nct_q = self.pd2.zeros()
        self.nct.add(self.nct_q, 1.0 / self.nspins)
        self.nct_G = self.pd2.ifft(self.nct_q)

    def interpolate_pseudo_density(self, comp_charge=None):
        """Interpolate pseudo density to fine grid."""
        if comp_charge is None:
            comp_charge, _Q_aL = self.calculate_multipole_moments()

        if self.nt_xg is None:
            self.nt_xg = self.finegd.empty(self.ncomponents)
            self.nt_sg = self.nt_xg[:self.nspins]
            self.nt_vg = self.nt_xg[self.nspins:]
            self.nt_Q = self.pd2.empty()

        self.nt_Q[:] = 0.0

        x = 0
        for nt_G, nt_g in zip(self.nt_xG, self.nt_xg):
            nt_g[:], nt_Q = self.pd2.interpolate(nt_G, self.pd3)
            if x < self.nspins:
                self.nt_Q += nt_Q
            x += 1

    def interpolate(self, in_xR, out_xR=None):
        """Interpolate array(s)."""
        if out_xR is None:
            out_xR = self.finegd.empty(in_xR.shape[:-3])

        a_xR = in_xR.reshape((-1,) + in_xR.shape[-3:])
        b_xR = out_xR.reshape((-1,) + out_xR.shape[-3:])

        for in_R, out_R in zip(a_xR, b_xR):
            out_R[:] = self.pd2.interpolate(in_R, self.pd3)[0]

        return out_xR

    distribute_and_interpolate = interpolate

    def calculate_pseudo_charge(self):
        self.rhot_q = self.pd3.zeros()
        Q_aL = self.Q.calculate(self.D_asp)
        self.ghat.add(self.rhot_q, Q_aL)
        self.map23.add_to2(self.rhot_q, self.nt_Q)
        self.background_charge.add_fourier_space_charge_to(self.pd3,
                                                           self.rhot_q)

    def get_pseudo_core_kinetic_energy_density_lfc(self):
        return PseudoCoreKineticEnergyDensityLFC(
            [[setup.tauct] for setup in self.setups], self.pd2)

    def calculate_dipole_moment(self):
        pd = self.pd3
        N_c = pd.tmp_Q.shape

        m0_q, m1_q, m2_q = [i_G == 0
                            for i_G in np.unravel_index(pd.Q_qG[0], N_c)]
        rhot_q = self.pd3.gather(self.rhot_q)
        if pd.comm.rank == 0:
            irhot_q = rhot_q.imag
            rhot_cs = [irhot_q[m1_q & m2_q],
                       irhot_q[m0_q & m2_q],
                       irhot_q[m0_q & m1_q]]
            d_c = [np.dot(rhot_s[1:], 1.0 / np.arange(1, len(rhot_s)))
                   for rhot_s in rhot_cs]
            d_v = -np.dot(d_c, pd.gd.cell_cv) / pi * pd.gd.dv
        else:
            d_v = np.empty(3)
        pd.comm.broadcast(d_v, 0)
        return d_v



import pytest
import numpy as np
from ase.build import molecule
from gpaw import GPAW
from gpaw.lcaotddft import LCAOTDDFT
from gpaw.lcaotddft.dipolemomentwriter import DipoleMomentWriter
from gpaw.mpi import world
from gpaw.test import equal


@pytest.mark.gllb
@pytest.mark.libxc
@pytest.mark.rttddft
def test_lcaotddft_fxc_vs_linearize(in_tmp_dir):
    atoms = molecule('SiH4')
    atoms.center(vacuum=4.0)

    # Ground-state calculation
    calc = GPAW(nbands=7, h=0.4,
                basis='dzp', mode='lcao',
                convergence={'density': 1e-8},
                xc='GLLBSC',
                symmetry={'point_group': False},
                txt='gs.out')
    atoms.calc = calc
    _ = atoms.get_potential_energy()
    calc.write('gs.gpw', mode='all')

    fxc = 'LDA'
    # Time-propagation calculation with fxc
    td_calc = LCAOTDDFT('gs.gpw', fxc=fxc, txt='td_fxc.out')
    DipoleMomentWriter(td_calc, 'dm_fxc.dat')
    td_calc.absorption_kick(np.ones(3) * 1e-5)
    td_calc.propagate(20, 4)

    # Time-propagation calculation with linearize_to_fxc()
    td_calc = LCAOTDDFT('gs.gpw', txt='td_lin.out')
    td_calc.linearize_to_xc(fxc)
    DipoleMomentWriter(td_calc, 'dm_lin.dat')
    td_calc.absorption_kick(np.ones(3) * 1e-5)
    td_calc.propagate(20, 4)

    # Test the equivalence
    world.barrier()
    ref = np.loadtxt('dm_fxc.dat').ravel()
    data = np.loadtxt('dm_lin.dat').ravel()

    tol = 1e-9
    equal(data, ref, tol)

    # Test the absolute values
    if 0:
        from gpaw.test import print_reference
        print_reference(data, 'ref', '%.12le')

    ref = [0.00000000e+00, 1.62932507e-15,
           -7.87693913e-15, -1.14696427e-14,
           -5.80989675e-15, 0.00000000e+00,
           3.62636797e-15, -3.29113009e-14,
           -3.51289578e-14, -3.67931628e-14,
           8.26827470e-01, -1.81172385e-15,
           6.17633458e-05, 6.17633458e-05,
           6.17633458e-05, 1.65365493e+00,
           -8.01473100e-16, 9.88572381e-05,
           9.88572381e-05, 9.88572381e-05,
           2.48048240e+00, -2.63296003e-15,
           1.04139540e-04, 1.04139540e-04,
           1.04139540e-04, 3.30730987e+00,
           2.29033034e-15, 8.81733367e-05,
           8.81733367e-05, 8.81733367e-05]

    print('result')
    print(data.tolist())

    tol = 1e-12
    equal(data, ref, tol)



# *** An error occurred in MPI_Recv
# *** on communicator MPI COMMUNICATOR 36 DUP FROM 28
# *** MPI_ERR_TRUNCATE: message truncated
# *** MPI_ERRORS_ARE_FATAL (your MPI job will now abort)

# works with 'sl_default': (2, 2, 32)

import pytest
from gpaw.mpi import world
from ase.build import fcc100, add_adsorbate
from gpaw import GPAW, ConvergenceError
from gpaw.utilities import compiled_with_sl

pytestmark = pytest.mark.skipif(world.size != 4,
                                reason='world.size != 4')


def test_parallel_scalapack_mpirecv_crash():
    assert world.size == 4

    slab = fcc100('Cu', size=(2, 2, 4))
    add_adsorbate(slab, 'O', 1.1, 'hollow')
    slab.center(vacuum=3.0, axis=2)

    kwargs = {}
    if compiled_with_sl():
        kwargs['parallel'] = {'domain': (1, 1, 4), 'sl_default': (2, 2, 64)}

    calc = GPAW(mode='lcao',
                kpts=(2, 2, 1),
                txt='-',
                maxiter=1,
                **kwargs)

    slab.calc = calc
    try:
        slab.get_potential_energy()
    except ConvergenceError:
        pass



import pytest
import numpy as np
from gpaw.grid_descriptor import GridDescriptor
from gpaw.spline import Spline
import gpaw.mpi as mpi
from gpaw.lfc import LocalizedFunctionsCollection as LFC
from gpaw.gpu import cupy as cp
from gpaw.core.atom_arrays import AtomArraysLayout


@pytest.mark.serial
@pytest.mark.gpu
@pytest.mark.parametrize('dtype', [float, complex])
def test_lfc(gpu, dtype):
    s = Spline(0, 1.0, [1.0, 0.5, 0.0])
    n = 40
    a = 8.0
    gd = GridDescriptor((n, n, n), (a, a, a), comm=mpi.world)

    c = LFC(gd, [[s], [s], [s]], dtype=dtype)
    c_gpu = LFC(gd, [[s], [s], [s]], dtype=dtype, xp=cp)
    c.set_positions([(0.5, 0.5, 0.25 + 0.25 * i) for i in [0, 1, 2]])
    c_gpu.set_positions([(0.5, 0.5, 0.25 + 0.25 * i) for i in [0, 1, 2]])

    P_ani = {}
    for a in c.my_atom_indices:
        P_ani[a] = np.ones((1,), dtype=dtype)

    P_gpu_ani = AtomArraysLayout([1, 1, 1], dtype=dtype, xp=cp).empty()
    P_gpu_ani.data[:] = 1.0

    b = gd.zeros(dtype=dtype)
    b_gpu = cp.zeros_like(b)
    q = {float: -1, complex: 0}[dtype]
    c.add(b, P_ani, q=q)
    c_gpu.add(b_gpu, P_gpu_ani, q=q)
    b_ref = b_gpu.get()

    assert b == pytest.approx(b_ref, abs=1e-12)

    c.integrate(b, P_ani, q=q)
    c_gpu.integrate(b_gpu, P_gpu_ani, q=q)

    for a, P_ni in P_ani.items():
        assert P_ni == pytest.approx(P_gpu_ani[a].get(), abs=1e-12)



import pytest
from ase.build import bulk
from gpaw import GPAW, FermiDirac, PW
from gpaw.mpi import serial_comm
from gpaw.xc.rpa import RPACorrelation
from gpaw.test import equal


@pytest.mark.rpa
@pytest.mark.response
def test_rpa_rpa_energy_Na(in_tmp_dir):
    blk = bulk('Na', 'bcc', a=4.23)

    ecut = 350
    calc = GPAW(mode=PW(ecut),
                basis='dzp',
                kpts={'size': (4, 4, 4), 'gamma': True},
                parallel={'domain': 1},
                txt='gs_occ_pw.txt',
                nbands=4,
                occupations=FermiDirac(0.01),
                setups={'Na': '1'})
    blk.calc = calc
    blk.get_potential_energy()
    calc.write('gs_occ_pw.gpw')

    calc = GPAW('gs_occ_pw.gpw', txt='gs_pw.txt', parallel={'band': 1})
    calc.diagonalize_full_hamiltonian(nbands=520)
    calc.write('gs_pw.gpw', 'all')

    ecut = 120
    calc = GPAW('gs_pw.gpw', communicator=serial_comm, txt=None)
    rpa = RPACorrelation(calc, txt='rpa_%s.txt' % ecut, ecut=[ecut])
    E = rpa.calculate()
    equal(E, -1.106, 0.005)



import pytest
from ase import Atoms
from gpaw import GPAW
from gpaw.xc.vdw import VDWFunctional


@pytest.mark.libxc
def test_vdw_quick():
    vdw = VDWFunctional('vdW-DF', verbose=1)
    L = 2.5
    a = Atoms(
        'H',
        cell=(L, L, L),
        pbc=True,
        calculator=GPAW(mode='fd', nbands=1))
    e = a.get_potential_energy()
    e2 = a.calc.get_xc_difference(vdw)
    print(e, e2)
    assert (vdw.shape == (24, 24, 24)).all()



import numpy as np
from gpaw.xc import XC
from gpaw.xc.gllb.contribution import Contribution


class C_XC(Contribution):
    def __init__(self, weight, functional):
        Contribution.__init__(self, weight)
        self.xc = XC(functional)

    def get_name(self):
        return 'XC'

    def get_desc(self):
        desc = self.xc.get_description()
        return self.xc.name if desc is None else desc

    def initialize(self, density, hamiltonian, wfs):
        Contribution.initialize(self, density, hamiltonian, wfs)
        self.vt_sg = self.finegd.empty(self.nspins)
        self.e_g = self.finegd.empty()

    def initialize_1d(self, ae):
        Contribution.initialize_1d(self, ae)
        self.v_g = np.zeros(self.ae.N)

    def calculate(self, e_g, n_sg, v_sg):
        self.e_g[:] = 0.0
        self.vt_sg[:] = 0.0
        self.xc.calculate(self.finegd, n_sg, self.vt_sg, self.e_g)
        v_sg += self.weight * self.vt_sg
        e_g += self.weight * self.e_g

    def calculate_energy_and_derivatives(self, setup, D_sp, H_sp, a,
                                         addcoredensity=True):
        E = self.xc.calculate_paw_correction(setup, D_sp, H_sp, True, a)
        E += setup.xc_correction.e_xc0
        return E

    def add_xc_potential_and_energy_1d(self, v_g):
        self.v_g[:] = 0.0
        Exc = self.xc.calculate_spherical(self.ae.rgd,
                                          self.ae.n.reshape((1, -1)),
                                          self.v_g.reshape((1, -1)))
        v_g += self.weight * self.v_g
        return self.weight * Exc

    def add_smooth_xc_potential_and_energy_1d(self, vt_g):
        self.v_g[:] = 0.0
        Exc = self.xc.calculate_spherical(self.ae.rgd,
                                          self.ae.nt.reshape((1, -1)),
                                          self.v_g.reshape((1, -1)))
        vt_g += self.weight * self.v_g
        return self.weight * Exc



from gpaw.xc.libxc import LibXC
from gpaw.xc.lda import LDA
from gpaw.xc.gga import GGA
from gpaw.xc.mgga import MGGA
from gpaw.xc.noncollinear import NonCollinearLDAKernel
from gpaw import libraries
from gpaw.xc.functional import XCFunctional


def xc_string_to_dict(string):
    """Convert XC specification string to dictionary.

    'name:key1=value1:...' -> {'name': <name>, key1: value1, ...}."""
    tokens = string.split(':')

    d = {'name': tokens[0]}
    for token in tokens[1:]:
        kw, val = token.split('=')
        # Convert value to int or float if possible
        try:
            val = int(val)
        except ValueError:
            try:
                val = float(val)
            except ValueError:
                pass
        d[kw] = val
    return d


def XC(kernel, parameters=None, atoms=None, collinear=True) -> XCFunctional:
    """Create XCFunctional object.

    kernel: XCKernel object, dict or str
        Kernel object or name of functional.
    parameters: ndarray
        Parameters for BEE functional.

    Recognized names are: LDA, PW91, PBE, revPBE, RPBE, BLYP, HCTH407,
    TPSS, M06-L, revTPSS, vdW-DF, vdW-DF2, EXX, PBE0, B3LYP, BEE,
    GLLBSC.  One can also use equivalent libxc names, for example
    GGA_X_PBE+GGA_C_PBE is equivalent to PBE, and LDA_X to the LDA exchange.
    In this way one has access to all the functionals defined in libxc.
    See xc_funcs.h for the complete list.  """

    if isinstance(kernel, str):
        kernel = xc_string_to_dict(kernel)

    kwargs = {}
    if isinstance(kernel, dict):
        kwargs = kernel.copy()
        name = kwargs.pop('name')
        backend = kwargs.pop('backend', None)

        if backend == 'libvdwxc' or name == 'vdW-DF-cx':
            # Must handle libvdwxc before old vdw implementation to override
            # behaviour for 'name'.  Also, cx is not implemented by the old
            # vdW module, so that always refers to libvdwxc.
            from gpaw.xc.libvdwxc import get_libvdwxc_functional
            return get_libvdwxc_functional(name=name, **kwargs)
        elif backend == 'ri':
            # Note: It is important that this if is before the next name is
            # HSExx, since otherwise PWHybrid would hijack the control flow.
            from gpaw.xc.ri import RI
            return RI(name, **kwargs)
        elif backend == 'pw' or name in ['HSE03', 'HSE06']:
            from gpaw.hybrids import HybridXC
            return HybridXC(name, **kwargs)  # type: ignore
        elif backend:
            raise ValueError(
                'A special backend for the XC functional was given, '
                'but not understood. Please check if there is a typo.')

        if name in ['vdW-DF', 'vdW-DF2', 'optPBE-vdW', 'optB88-vdW',
                    'C09-vdW', 'mBEEF-vdW', 'BEEF-vdW']:
            from gpaw.xc.vdw import VDWFunctional
            return VDWFunctional(name, **kwargs)
        elif name in ['EXX', 'PBE0', 'B3LYP',
                      'CAMY-BLYP', 'CAMY-B3LYP', 'LCY-BLYP', 'LCY-PBE']:
            from gpaw.xc.hybrid import HybridXC as OldHybridXC
            return OldHybridXC(name, **kwargs)  # type: ignore
        elif name.startswith('LCY-') or name.startswith('CAMY-'):
            parts = name.split('(')
            from gpaw.xc.hybrid import HybridXC as OldHybridXC
            return OldHybridXC(parts[0],
                               omega=float(parts[1][:-1]))
        elif name == 'BEE2':
            from gpaw.xc.bee import BEE2
            kernel = BEE2(parameters)
        elif name.startswith('GLLB'):
            from gpaw.xc.gllb.nonlocalfunctionalfactory import (
                get_nonlocal_functional)
            xc = get_nonlocal_functional(name, **kwargs)
            return xc
        elif name == 'LB94':
            from gpaw.xc.lb94 import LB94
            kernel = LB94()
        elif name == 'TB09':
            from gpaw.xc.tb09 import TB09
            return TB09(**kwargs)
        elif name.endswith('PZ-SIC'):
            from gpaw.xc.sic import SIC
            return SIC(xc=name[:-7], **kwargs)
        elif name in {'TPSS', 'revTPSS', 'M06-L'}:
            assert libraries['libxc'], 'Please compile with libxc'
            from gpaw.xc.kernel import XCKernel
            kernel = XCKernel(name)
        elif name in {'LDA', 'PBE', 'revPBE', 'RPBE', 'PW91'}:
            from gpaw.xc.kernel import XCKernel
            kernel = XCKernel(name)
        elif name.startswith('old'):
            from gpaw.xc.kernel import XCKernel
            kernel = XCKernel(name[3:])
        elif name == 'PPLDA':
            from gpaw.xc.lda import PurePythonLDAKernel
            kernel = PurePythonLDAKernel()
        elif name in ['pyPBE', 'pyPBEsol', 'pyRPBE', 'pyzvPBEsol']:
            from gpaw.xc.gga import PurePythonGGAKernel
            kernel = PurePythonGGAKernel(name)
        elif name == '2D-MGGA':
            from gpaw.xc.mgga import PurePython2DMGGAKernel
            kernel = PurePython2DMGGAKernel(name, parameters)
        elif name[0].isdigit():
            from gpaw.xc.parametrizedxc import ParametrizedKernel
            kernel = ParametrizedKernel(name)
        elif name == 'null':
            from gpaw.xc.kernel import XCNull
            kernel = XCNull()
        elif name == 'QNA':
            from gpaw.xc.qna import QNA
            return QNA(atoms, kernel['parameters'], kernel['setup_name'],
                       alpha=kernel['alpha'], stencil=kwargs.get('stencil', 2))
        else:
            kernel = LibXC(name)

    if kernel.type == 'LDA':
        if not collinear:
            kernel = NonCollinearLDAKernel(kernel)
        return LDA(kernel, **kwargs)

    elif kernel.type == 'GGA':
        return GGA(kernel, **kwargs)
    else:
        return MGGA(kernel, **kwargs)



"""GPAW wannier example for ethylene corresponding to the ASE Wannier
tutorial.
"""
import pytest
import numpy as np
from ase import Atoms

from gpaw import GPAW
from gpaw.mpi import size
from gpaw.wannier.overlaps import calculate_overlaps
from gpaw.wannier.edmiston_ruedenberg import localize

pytestmark = pytest.mark.ci


@pytest.fixture(scope='module')
def ethylene():
    a = 6.0  # Size of unit cell (Angstrom)

    mol = Atoms('H2C2H2',
                [(-1.235, -0.936, 0),
                 (-1.235, 0.936, 0),
                 (-0.660, 0.000, 0),
                 (0.660, 0.000, 0),
                 (1.235, -0.936, 0),
                 (1.235, 0.936, 0)],
                cell=(a, a, a),
                pbc=True)
    mol.center()

    mol.calc = GPAW(mode='fd',
                    txt=None,
                    nbands=8,
                    gpts=(32, 32, 32),
                    convergence={'eigenstates': 3.3e-5})
    mol.get_potential_energy()
    return mol


@pytest.mark.skipif(size > 1, reason='Not parallelized')
def test_ethylene_energy(ethylene):
    e = ethylene.get_potential_energy()
    assert e == pytest.approx(-33.328, abs=0.002)


def check(calc):
    wannier = localize(calculate_overlaps(calc, n1=0, n2=6, nwannier=6))

    centers = wannier.centers
    print(centers)
    expected = [[1.950, 2.376, 3.000],
                [1.950, 3.624, 3.000],
                [3.000, 3.000, 2.671],
                [3.000, 3.000, 3.329],
                [4.050, 2.376, 3.000],
                [4.050, 3.624, 3.000]]
    assert wannier.value == pytest.approx(13.7995, abs=0.016)
    for center in centers:
        i = 0
        while np.sum((expected[i] - center)**2) > 0.01:
            i += 1
            if i == len(expected):
                raise RuntimeError('Correct center not found')
        expected.pop(i)


@pytest.mark.skipif(size > 1, reason='Not parallelized')
def test_wannier_centers(ethylene):
    check(ethylene.calc)


@pytest.mark.skipif(size > 1, reason='Not parallelized')
def test_wannier_centers_gpw(ethylene, in_tmp_dir):
    ethylene.calc.write('ethylene.gpw', 'all')
    check(GPAW('ethylene.gpw', txt=None))



"""Developer quiz."""
from __future__ import annotations

import contextlib
import io
from dataclasses import dataclass
from random import shuffle
from textwrap import wrap

with contextlib.redirect_stdout(io.StringIO()):
    from this import d


@dataclass
class Question:
    text: str
    choises: list[str] | None = None
    solutions: list[str] | None = None
    shuffle: bool = True
    extra_check: bool = False

    def ask(self) -> bool:
        """Ask question and check result."""
        print()
        text = self.text.rstrip('?') + '?'
        for line in text.splitlines():
            print('\n'.join(wrap(line)))
        if self.choises:
            N = len(self.choises)
            indices = list(range(N))
            if self.shuffle:
                shuffle(indices)
            i0 = len(self.text) - len(text)
            n0 = indices.index(i0) + 1
            for n, i in enumerate(indices, start=1):
                print(f'{n}: {self.choises[i]}')
            answer = input(f'[1-{N}]: ')
            if answer != str(n0):
                return False
            if self.extra_check:
                return Question(text='Really?', choises=['yes', 'no']).ask()
            return True

        answer = input()
        return self.solutions is None or encode(answer) in self.solutions


def encode(text: str) -> str:
    """See "The Zen of Python" (this.py)."""
    return ''.join(d.get(c, c) for c in text)


dev = 'https://wiki.fysik.dtu.dk/gpaw/devel'

questions = [
    Question(text='What, ... is your name?'),
    Question(text='What, ... is your quest?',
             choises=['To become a GPAW-developer']),
    Question(text='What, ... is your favourite programming language?',
             choises=['Python', 'Other'],
             shuffle=False),
    Question('What, ... is the air-speed velocity of an unladen swallow???',
             choises=['10 m/s',
                      '20 miles an hour',
                      'depends on the kind of swallow (African or European)']),
    Question(text='Here is some important developer information:\n\n'
             f'  {dev}/workflow.html\n'
             f'  {dev}/testing.html\n'
             f'  {dev}/writing_documentation.html\n\n'
             'Did you read all of these three pages??',
             choises=['no', 'yes'],
             extra_check=True),
    Question(text='What framework does GPAW use for its test suite??',
             choises=['home-made system',
                      'pytest',
                      'unittest']),
    Question(text='What is the name of the environment variable that points '
             'at the folder(s) where PAW dataset files are stores?',
             solutions=['TCNJ_FRGHC_CNGU', '$TCNJ_FRGHC_CNGU'])]


def main() -> None:
    """Ask questions."""
    print('To become a gpaw developer, you must answer these '
          f'{len(questions)} questions ...')
    shuffle(questions)
    for question in questions:
        ok = question.ask()
        if ok:
            print('\033[92mCorrect!\033[0m')
        else:
            print('\033[91mAuuuuuuuugh!\033[0m')
            return
    print('\nRight. Off you go.')


if __name__ == '__main__':
    main()



"""This module provides a class for assembling the dynamical matrix."""

__all__ = ["DynamicalMatrix"]

import os
import pickle

import numpy as np
import numpy.fft as fft

from gpaw import debug
from gpaw.mpi import serial_comm
from gpaw.utilities import unpack


class DynamicalMatrix:
    """Class for assembling the dynamical matrix from first-order responses.

    The second order derivative of the total energy with respect to atomic
    displacements (for periodic systems collective atomic displacemnts
    characterized by a q-vector) can be obtained from an expression involving
    the first-order derivatives of the density and the wave-functions.

    Each of the various contributions to the second order derivative of the
    total energy are implemented in separate functions.

    """

    def __init__(self, atoms, kd, dtype=float):
        """Inititialize class with a list of atoms.

        Parameters
        ----------
        atoms: Atoms
            List of atoms for the system.
        kd: KPointDescriptor
            Descriptor for the q-vector grid.

        """

        # Store useful objects
        self.atoms = atoms
        self.kd = kd
        self.dtype = dtype
        self.masses = atoms.get_masses()
        # Array with inverse sqrt of masses repeated to match shape of mode
        # arrays
        self.m_inv_av = None

        # String template for files with force constants
        self.name = None
        # Path to directory with force constant files
        self.path = None

        # List of atomic indices to be included (default is all atoms)
        self.indices = range(len(self.atoms))

        # Index of the gamma point -- for the acoustic sum-rule
        self.gamma_index = None

        if self.kd.gamma:
            self.gamma_index = 0
            assert dtype == float
        else:
            for k, k_c in enumerate(self.kd.ibzk_kc):
                if np.all(k_c == 0.):
                    self.gamma_index = k

        assert self.gamma_index is not None

        # Matrix of force constants -- dict of dicts in atomic indices
        # Only q-vectors in the irreducible BZ stored here
        self.C_qaavv = [dict([(a, dict([(a_, np.zeros((3, 3), dtype=dtype))
                                        for a_ in self.indices]))
                              for a in self.indices])
                        for q in range(self.kd.nibzkpts)]

        # Force constants and dynamical matrix attributes
        # Irreducible zone -- list with array entrances C_avav
        self.C_q = None
        # Full BZ (ndarray)
        self.D_k = None

        self.assembled = False

    def __getstate__(self):
        """Method used to pickle an instance of ``DynamicalMatrix``.

        Bound method attributes cannot be pickled and must therefore be deleted
        before an instance is dumped to file.

        """

        # Get state of object and take care of troublesome attributes
        state = dict(self.__dict__)
        state['kd'].__dict__['comm'] = serial_comm

        return state

    def __setstate__(self, state):
        """Method used to unpickle an instance of ``DynamicalMatrix``."""

        self.__dict__.update(state)

    def get_mass_array(self):
        """Return inverse sqrt of masses (matches shape of mode array)."""

        assert self.m_inv_av is not None
        return self.m_inv_av

    def get_indices(self):
        """Return indices of included atoms."""

        return self.indices

    def set_indices(self, indices):
        """Set indices of atoms to be included in the calculation."""

        self.indices = indices
        self.C_qaavv = [dict([(a,
                               dict([(a_, np.zeros((3, 3), dtype=self.dtype))
                                     for a_ in self.indices]))
                              for a in self.indices])
                        for q in range(self.kd.nibzkpts)]

    def set_name_and_path(self, name, path):
        """Set name and path of the force constant files.

        name: str
            Base name template for the files which the elements of the matrix
            of force constants are written to.
        path: str
            Path specifying the directory with the files.

        """

        self.name = name
        self.path = path

    def write(self, fd, q, a, v):
        """Write force constants for specified indices to file."""

        C_qav_a = [self.C_qaavv[q][a][a_][v] for a_ in self.indices]
        pickle.dump(C_qav_a, fd)

    def read(self):
        """Read the force constants from files."""

        assert self.name is not None
        assert self.path is not None

        for q in range(self.kd.nibzkpts):
            for a in self.indices:
                for v in [0, 1, 2]:
                    filename = self.name % (q, a, v)
                    try:
                        fd = open(os.path.join(self.path, filename), 'rb')
                    except EOFError:
                        print(("Redo file %s "
                               % os.path.join(self.path, filename)))
                    C_qav_a = pickle.load(fd)
                    fd.close()
                    for a_ in self.indices:
                        self.C_qaavv[q][a][a_][v] = C_qav_a[a_]

    def assemble(self, acoustic=True):
        """Assemble dynamical matrix from the force constants attribute.

        The elements of the dynamical matrix are given by::

            D_ij(q) = 1/(M_i + M_j) * C_ij(q) ,

        where i and j are collective atomic and cartesian indices.

        During the assembly, various symmetries of the dynamical matrix are
        enforced::

            1) Hermiticity
            2) Acoustic sum-rule
            3) D(q) = D*(-q)

        Parameters
        ----------
        acoustic: bool
            When True, the diagonal of the matrix of force constants is
            corrected to ensure that the acoustic sum-rule is fulfilled.

        """

        # Read force constants from files
        self.read()

        # Number of atoms included
        N = len(self.indices)

        # Assemble matrix of force constants
        self.C_q = []
        for q, C_aavv in enumerate(self.C_qaavv):

            C_avav = np.zeros((3 * N, 3 * N), dtype=self.dtype)

            for i, a in enumerate(self.indices):
                for j, a_ in enumerate(self.indices):
                    C_avav[3 * i:3 * i + 3, 3 * j:3 * j + 3] += C_aavv[a][a_]

            self.C_q.append(C_avav)

        # XXX Figure out in which order the corrections should be done
        # Make C(q) Hermitian
        for C in self.C_q:
            C *= 0.5
            C += C.conj().T

        # Get matrix of force constants in the Gamma-point (is real!)
        C_gamma = self.C_q[self.gamma_index].real
        # Make Gamma-component real
        self.C_q[self.gamma_index] = C_gamma.copy()

        # Apply acoustic sum-rule if requested
        if acoustic:
            # Correct atomic diagonal for each q-vector
            for C in self.C_q:
                for a in range(N):
                    for a_ in range(N):
                        C[3 * a:3 * a + 3, 3 * a:3 * a + 3] -= \
                            C_gamma[3 * a:3 * a + 3, 3 * a:3 * a_ + 3]

            # Check sum-rule for Gamma-component in debug mode
            if debug:
                C = self.C_q[self.gamma_index]
                assert np.all(np.sum(C.reshape((3 * N, N, 3)), axis=1) < 1e-15)

        # Move this bit to an ``unfold`` member function
        # XXX Time-reversal symmetry
        C_q = np.asarray(self.C_q)
        if self.kd.nibzkpts != self.kd.nbzkpts:
            self.D_k = np.concatenate((C_q[:0:-1].conjugate(), C_q))
        else:
            self.D_k = 0.5 * C_q
            self.D_k += self.D_k[::-1].conjugate()

        # Mass prefactor for the dynamical matrix
        self.m_inv_av = np.repeat(self.masses[self.indices]**-0.5, 3)
        M_avav = self.m_inv_av[:, np.newaxis] * self.m_inv_av

        for C in self.D_k:
            C *= M_avav

        self.assembled = True

    def real_space(self):
        """Fourier transform the dynamical matrix to real-space."""

        if not self.assembled:
            self.assemble()

        # Shape of q-point grid
        N_c = tuple(self.kd.N_c)

        # Reshape before Fourier transforming
        shape = self.D_k.shape
        Dq_lmn = self.D_k.reshape(N_c + shape[1:])
        DR_lmn = fft.ifftn(fft.ifftshift(Dq_lmn, axes=(0, 1, 2)),
                           axes=(0, 1, 2))

        if debug:
            # Check that D_R is real enough
            assert np.all(DR_lmn.imag < 1e-8)

        DR_lmn = DR_lmn.real

        # Corresponding R_m vectors in units of the lattice vectors
        R_cm = np.indices(N_c).reshape(3, -1)
        N1_c = np.array(N_c)[:, np.newaxis]
        R_cm += N1_c // 2
        R_cm %= N1_c
        R_cm -= N1_c // 2
        R_clmn = R_cm.reshape((3,) + N_c)

        return DR_lmn, R_clmn

    def fourier_interpolate(self, N_c):
        """Fourier interpolate dynamical matrix onto a finer q-vector grid."""

        raise NotImplementedError

    def calculate_row(self, perturbation, response_calc):
        """Calculate row of force constant matrix from first-order derivatives.

        Parameters
        ----------
        perturbation: PhononPerturbation
            The perturbation which holds the derivative of the
            pseudo-potential.
        response_calc: ResponseCalculator
            Calculator with the corresponding derivatives of the density and
            the wave-functions.

        """

        self.density_derivative(perturbation, response_calc)
        self.wfs_derivative(perturbation, response_calc)

    def density_ground_state(self, calc):
        """Contributions involving ground-state density.

        These terms contains second-order derivaties of the localized functions
        ghat and vbar. They are therefore diagonal in the atomic indices.

        """

        # Use the GS LFC's to integrate with the ground-state quantities !
        ghat = calc.density.ghat
        vbar = calc.hamiltonian.vbar
        # Compensation charge coefficients
        Q_aL = calc.density.Q_aL

        # Integral of Hartree potential times the second derivative of ghat
        vH_g = calc.hamiltonian.vHt_g
        d2ghat_aLvv = dict([(atom.index, np.zeros((3, 3)))
                            for atom in self.atoms])
        ghat.second_derivative(vH_g, d2ghat_aLvv)

        # Integral of electron density times the second derivative of vbar
        nt_g = calc.density.nt_g
        d2vbar_avv = dict([(atom.index, np.zeros((3, 3)))
                           for atom in self.atoms])
        vbar.second_derivative(nt_g, d2vbar_avv)

        # Matrix of force constants to be updated; q=-1 for Gamma calculation!
        for C_aavv in self.C_qaavv:

            for a in self.indices:
                # XXX: HGH has only one ghat pr atoms -> generalize when
                # implementing PAW
                C_aavv[a][a] += d2ghat_aLvv[a] * Q_aL[a]
                C_aavv[a][a] += d2vbar_avv[a]

    def wfs_ground_state(self, calc, response_calc):
        """Ground state contributions from the non-local potential."""

        # Projector functions
        pt = calc.wfs.pt
        # Projector coefficients
        dH_asp = calc.hamiltonian.dH_asp

        # K-point
        kpt_u = response_calc.wfs.kpt_u
        nbands = response_calc.nbands

        for kpt in kpt_u:

            # Index of k
            P_ani = kpt.P_ani
            dP_aniv = kpt.dP_aniv

            # Wave functions
            psit_nG = kpt.psit_nG
            # psit1_nG = kpt.psit1_nG

            # Calculate d2P_anivv coefficients
            # d2P_anivv = self.calculate_d2P_anivv()
            d2P_anivv = dict(
                [(atom.index,
                  np.zeros(
                      (nbands, pt.get_function_count(atom.index), 3, 3)))
                 for atom in self.atoms])
            # XXX Temp dict, second_derivative method only takes a_G array
            # -- no extra dims
            d2P_avv = dict([(atom.index, np.zeros((3, 3)))
                            for atom in self.atoms])

            for n in range(nbands):
                pt.second_derivative(psit_nG[n], d2P_avv)
                # Insert in other dict
                for atom in self.atoms:
                    a = atom.index
                    d2P_anivv[a][n, 0] = d2P_avv[a]

            for a in self.indices:

                H_ii = unpack(dH_asp[a][0])
                P_ni = P_ani[a]
                dP_niv = -1 * dP_aniv[a]
                d2P_nivv = d2P_anivv[a]

                # Term with second-order derivative of projector
                HP_ni = np.dot(P_ni, H_ii)
                d2PHP_nvv = (d2P_nivv.conj() *
                             HP_ni[:, :, np.newaxis, np.newaxis]).sum(1)
                d2PHP_nvv *= kpt.weight
                A_vv = d2PHP_nvv.sum(0)

                # Term with first-order derivative of the projectors
                HdP_inv = np.dot(H_ii, dP_niv.conj())
                HdP_niv = np.swapaxes(HdP_inv, 0, 1)
                HdP_niv *= kpt.weight

                B_vv = (dP_niv[:, :, np.newaxis, :] *
                        HdP_niv[:, :, :, np.newaxis]).sum(0).sum(0)

                for C_aavv in self.C_qaavv:
                    C_aavv[a][a] += (A_vv + B_vv) + (A_vv + B_vv).conj()

    def core_corrections(self):
        """Contribution from the derivative of the core density."""

        raise NotImplementedError

    def density_derivative(self, perturbation, response_calc):
        """Contributions involving the first-order density derivative."""

        # Get attributes from the phononperturbation
        a = perturbation.a
        v = perturbation.v
        # XXX careful here, Gamma calculation has q=-1
        q = perturbation.q

        # Matrix of force constants to be updated; q=-1 for Gamma calculation!
        C_aavv = self.C_qaavv[q]

        # Localized functions
        ghat = perturbation.ghat
        vbar = perturbation.vbar
        # Compensation charge coefficients
        Q_aL = perturbation.Q_aL

        # Density derivative
        nt1_g = response_calc.nt1_g

        # Hartree potential derivative including compensation charges
        vH1_g = response_calc.vH1_g.copy()
        vH1_g += perturbation.vghat1_g

        # Integral of Hartree potential derivative times ghat derivative
        dghat_aLv = ghat.dict(derivative=True)
        # Integral of density derivative times vbar derivative
        dvbar_av = vbar.dict(derivative=True)

        # Evaluate integrals
        ghat.derivative(vH1_g, dghat_aLv, q=q)
        vbar.derivative(nt1_g, dvbar_av, q=q)

        # Add to force constant matrix attribute
        for a_ in self.indices:
            # Minus sign comes from lfc member function derivative
            C_aavv[a][a_][v] -= np.dot(Q_aL[a_], dghat_aLv[a_])
            C_aavv[a][a_][v] -= dvbar_av[a_][0]

    def wfs_derivative(self, perturbation, response_calc):
        """Contributions from the non-local part of the PAW potential."""

        # Get attributes from the phononperturbation
        a = perturbation.a
        v = perturbation.v
        q = perturbation.q

        # Matrix of force constants to be updated
        C_aavv = self.C_qaavv[q]

        # Projector functions
        pt = response_calc.wfs.pt
        # Projector coefficients
        dH_asp = perturbation.dH_asp

        # K-point
        kpt_u = response_calc.wfs.kpt_u
        nbands = response_calc.nbands

        # Get k+q indices
        if perturbation.has_q():
            q_c = perturbation.get_q()
            kplusq_k = response_calc.wfs.kd.find_k_plus_q(q_c)
        else:
            kplusq_k = range(len(kpt_u))

        for kpt in kpt_u:

            # Indices of k and k+q
            k = kpt.k
            kplusq = kplusq_k[k]

            # Projector coefficients
            P_ani = kpt.P_ani
            dP_aniv = kpt.dP_aniv

            # psit_nG = kpt.psit_nG
            psit1_nG = kpt.psit1_nG

            # Overlap between wave-function derivative and projectors
            Pdpsi_ani = pt.dict(shape=nbands, zero=True)
            pt.integrate(psit1_nG, Pdpsi_ani, q=kplusq)
            # Overlap between wave-function derivative and derivative of
            # projectors
            dPdpsi_aniv = pt.dict(shape=nbands, derivative=True)
            pt.derivative(psit1_nG, dPdpsi_aniv, q=kplusq)

            for a_ in self.indices:
                # Coefficients from atom a
                Pdpsi_ni = Pdpsi_ani[a]
                dPdpsi_niv = -1 * dPdpsi_aniv[a]
                # Coefficients from atom a_
                H_ii = unpack(dH_asp[a_][0])
                P_ni = P_ani[a_]
                dP_niv = -1 * dP_aniv[a_]

                # Term with dPdpsi and P coefficients
                HP_ni = np.dot(P_ni, H_ii)
                dPdpsiHP_nv = (dPdpsi_niv.conj() *
                               HP_ni[:, :, np.newaxis]).sum(1)
                dPdpsiHP_nv *= kpt.weight
                A_v = dPdpsiHP_nv.sum(0)

                # Term with dP and Pdpsi coefficients
                HPdpsi_ni = np.dot(Pdpsi_ni.conj(), H_ii)
                dPHPdpsi_nv = (dP_niv * HPdpsi_ni[:, :, np.newaxis]).sum(1)
                dPHPdpsi_nv *= kpt.weight
                B_v = dPHPdpsi_nv.sum(0)

                # Factor of 2 from time-reversal symmetry
                C_aavv[a][a_][v] += 2 * (A_v + B_v)



"""This module implements a simple wave-function class."""

from math import pi
import numpy as np

from gpaw.lfc import LocalizedFunctionsCollection as LFC
from gpaw.dfpt.kpointcontainer import KPointContainer


class WaveFunctions:
    """Class for wave-function related stuff (e.g. projectors)."""

    def __init__(self, nbands, kpt_u, setups, kd, gd, dtype=float):
        """Store and initialize required attributes.

        Parameters
        ----------
        nbands: int
            Number of occupied bands.
        kpt_u: list of KPoints
            List of KPoint instances from a ground-state calculation (i.e. the
            attribute ``calc.wfs.kpt_u``).
        setups: Setups
            LocalizedFunctionsCollection setups.
        kd: KPointDescriptor
            K-point and symmetry related stuff.
        gd: GridDescriptor
            Descriptor for the coarse grid.
        dtype: dtype
            This is the ``dtype`` for the wave-function derivatives (same as
            the ``dtype`` for the ground-state wave-functions).

        """

        self.dtype = dtype
        # K-point related attributes
        self.kd = kd
        # Number of occupied bands
        self.nbands = nbands
        # Projectors
        self.pt = LFC(gd, [setup.pt_j for setup in setups], kd,
                      dtype=self.dtype)
        # Store grid
        self.gd = gd

        # Unfold the irreducible BZ to the full BZ
        # List of KPointContainers for the k-points in the full BZ
        self.kpt_u = []

        # No symmetries or only time-reversal symmetry used
        assert not kd.symmetry.point_group
        if not kd.symmetry.time_reversal:
            # For now, time-reversal symmetry not allowed
            assert len(kpt_u) == kd.nbzkpts

            for k in range(kd.nbzkpts):
                kpt_ = kpt_u[k]

                psit_nG = gd.empty(nbands, dtype=self.dtype)

                for n, psit_G in enumerate(psit_nG):
                    psit_G[:] = kpt_.psit_nG[n]
                    # psit_0 = psit_G[0, 0, 0]
                    # psit_G *= psit_0.conj() / (abs(psit_0))

                # Strip off KPoint attributes and store in the KPointContainer
                # Note, only the occupied GS wave-functions are retained here !
                kpt = KPointContainer(weight=kpt_.weight,
                                      k=kpt_.k,
                                      s=kpt_.s,
                                      phase_cd=kpt_.phase_cd,
                                      eps_n=kpt_.eps_n[:nbands],
                                      psit_nG=psit_nG,
                                      psit1_nG=None,
                                      P_ani=None,
                                      dP_aniv=None)
                # q=kpt.q,
                # f_n=kpt.f_n[:nbands])

                self.kpt_u.append(kpt)

        else:
            assert len(kpt_u) == kd.nibzkpts

            for k, k_c in enumerate(kd.bzk_kc):

                # Index of symmetry related point in the irreducible BZ
                ik = kd.bz2ibz_k[k]
                # Index of point group operation
                s = kd.sym_k[k]
                # Time-reversal symmetry used
                time_reversal = kd.time_reversal_k[k]

                # Coordinates of symmetry related point in the irreducible BZ
                ik_c = kd.ibzk_kc[ik]
                # Point group operation
                op_cc = kd.symmetry.op_scc[s]

                # KPoint from ground-state calculation
                kpt_ = kpt_u[ik]
                weight = 1. / kd.nbzkpts * (2 - kpt_.s)
                phase_cd = np.exp(2j * pi * gd.sdisp_cd * k_c[:, np.newaxis])

                psit_nG = gd.empty(nbands, dtype=self.dtype)

                for n, psit_G in enumerate(psit_nG):
                    # XXX Seems to corrupt my memory somehow ???
                    psit_G[:] = kd.symmetry.symmetrize_wavefunction(
                        kpt_.psit_nG[n], ik_c, k_c, op_cc, time_reversal)
                    # Choose gauge
                    # psit_0 = psit_G[0, 0, 0]
                    # psit_G *= psit_0.conj() / (abs(psit_0))

                kpt = KPointContainer(weight=weight,
                                      k=k,
                                      s=kpt_.s,
                                      phase_cd=phase_cd,
                                      eps_n=kpt_.eps_n[:nbands],
                                      psit_nG=psit_nG,
                                      psit1_nG=None,
                                      P_ani=None,
                                      dP_aniv=None)

                self.kpt_u.append(kpt)

    def initialize(self, spos_ac):
        """Initialize projectors according to the ``gamma`` attribute."""

        # Set positions on LFC's
        self.pt.set_positions(spos_ac)

        # Calculate projector coefficients for the GS wave-functions
        self.calculate_projector_coef()

    def reset(self):
        """Make fresh arrays for wave-function derivatives."""

        for kpt in self.kpt_u:
            kpt.psit1_nG = self.gd.zeros(n=self.nbands, dtype=self.dtype)

    def calculate_projector_coef(self):
        """Coefficients for the derivative of the non-local part of the PP.

        Parameters
        ----------
        k: int
            Index of the k-point of the Bloch state on which the non-local
            potential operates on.

        The calculated coefficients are the following (except for an overall
        sign of -1; see ``derivative`` member function of class ``LFC``):

        1. Coefficients from the projector functions::

                        /      a
               P_ani =  | dG  p (G) Psi (G)  ,
                        /      i       n

        2. Coefficients from the derivative of the projector functions::

                          /      a
               dP_aniv =  | dG dp  (G) Psi (G)  ,
                          /      iv       n

        where::

                 a        d       a
               dp  (G) =  ---  Phi (G) .
                 iv         a     i
                          dR

        """

        n = self.nbands

        for kpt in self.kpt_u:

            # K-point index and wave-functions
            k = kpt.k
            psit_nG = kpt.psit_nG

            # Integration dicts
            P_ani = self.pt.dict(shape=n)
            dP_aniv = self.pt.dict(shape=n, derivative=True)

            # 1) Integrate with projectors
            self.pt.integrate(psit_nG, P_ani, q=k)
            kpt.P_ani = P_ani

            # 2) Integrate with derivative of projectors
            self.pt.derivative(psit_nG, dP_aniv, q=k)
            kpt.dP_aniv = dP_aniv



# Copyright (C) 2003  CAMP
# Please see the accompanying LICENSE file for further information.

import warnings
from math import pi

import numpy as np
from numpy.fft import fftn, ifftn, fft2, ifft2, rfft2, irfft2, fft, ifft
from scipy.fftpack import dst as scipydst


from gpaw import PoissonConvergenceError
from gpaw.dipole_correction import DipoleCorrection, dipole_correction
from gpaw.domain import decompose_domain
from gpaw.fd_operators import Laplace, LaplaceA, LaplaceB
from gpaw.transformers import Transformer
from gpaw.utilities.blas import axpy
from gpaw.utilities.gauss import Gaussian
from gpaw.utilities.grid import grid2grid
from gpaw.utilities.tools import construct_reciprocal
from gpaw.utilities.timing import NullTimer

POISSON_GRID_WARNING = """Grid unsuitable for FDPoissonSolver!

Consider using FastPoissonSolver instead.

The FDPoissonSolver does not have sufficient multigrid levels for good
performance and will converge inefficiently if at all, or yield wrong
results.

You may need to manually specify a grid such that the number of points
along each direction is divisible by a high power of 2, such as 8, 16,
or 32 depending on system size; examples:

  GPAW(gpts=(32, 32, 288))

or

  from gpaw.utilities import h2gpts
  GPAW(gpts=h2gpts(0.2, atoms.get_cell(), idiv=16))

Parallelizing over very small domains can also undesirably limit the
number of multigrid levels even if the total number of grid points
is divisible by a high power of 2."""


def create_poisson_solver(name='fast', **kwargs):
    if isinstance(name, _PoissonSolver):
        return name
    elif isinstance(name, dict):
        kwargs.update(name)
        return create_poisson_solver(**kwargs)
    elif name == 'fft':
        return FFTPoissonSolver(**kwargs)
    elif name == 'fdtd':
        from gpaw.fdtd.poisson_fdtd import FDTDPoissonSolver
        return FDTDPoissonSolver(**kwargs)
    elif name == 'fd':
        return FDPoissonSolverWrapper(**kwargs)
    elif name == 'fast':
        return FastPoissonSolver(**kwargs)
    elif name == 'ExtraVacuumPoissonSolver':
        from gpaw.poisson_extravacuum import ExtraVacuumPoissonSolver
        return ExtraVacuumPoissonSolver(**kwargs)
    elif name == 'MomentCorrectionPoissonSolver':
        from gpaw.poisson_moment import MomentCorrectionPoissonSolver
        return MomentCorrectionPoissonSolver(**kwargs)
    elif name == 'nointeraction':
        return NoInteractionPoissonSolver()
    else:
        raise ValueError('Unknown poisson solver: %s' % name)


def PoissonSolver(name='fast', dipolelayer=None, **kwargs):
    p = create_poisson_solver(name=name, **kwargs)
    if dipolelayer is not None:
        p = DipoleCorrection(p, dipolelayer)
    return p


def FDPoissonSolverWrapper(dipolelayer=None, **kwargs):
    if dipolelayer is not None:
        return DipoleCorrection(FDPoissonSolver(**kwargs), dipolelayer)
    return FDPoissonSolver(**kwargs)


class _PoissonSolver(object):
    """Abstract PoissonSolver class

       This class defines an interface and a common ancestor
       for various PoissonSolver implementations (including wrappers)."""
    def __init__(self):
        object.__init__(self)

    def set_grid_descriptor(self, gd):
        raise NotImplementedError()

    def solve(self):
        raise NotImplementedError()

    def todict(self):
        raise NotImplementedError(self.__class__.__name__)

    def get_description(self):
        return self.__class__.__name__

    def estimate_memory(self, mem):
        raise NotImplementedError()


class BasePoissonSolver(_PoissonSolver):
    def __init__(self, *, remove_moment=None, use_charge_center=False,
                 metallic_electrodes=False, eps=None):

        if eps is not None:
            warnings.warn(
                "Please do not specify the eps parameter "
                f"for {self.__class__.__name__}. "
                "The parameter doesn't do anything for this solver "
                "and defining it will throw an error in the future.",
                FutureWarning)

        if remove_moment is not None:
            warnings.warn(
                "Please do not specify the remove_moment parameter "
                f"for {self.__class__.__name__}. "
                "The remove moment functionality is deprecated in this solver "
                "and will throw an error in the future. Instead "
                "use the MomentCorrectionPoissonSolver as a wrapper to "
                f"{self.__class__.__name__}.",
                FutureWarning)

        # metallic electrodes: mirror image method to allow calculation of
        # charged, partly periodic systems
        self.gd = None
        self.remove_moment = remove_moment
        self.use_charge_center = use_charge_center
        self.metallic_electrodes = metallic_electrodes
        assert self.metallic_electrodes in [False, None, 'single', 'both']

    def todict(self):
        d = {'name': 'basepoisson'}
        if self.remove_moment:
            d['remove_moment'] = self.remove_moment
        if self.use_charge_center:
            d['use_charge_center'] = self.use_charge_center
        if self.metallic_electrodes:
            d['metallic_electrodes'] = self.metallic_electrodes

        return d

    def get_description(self):
        # The idea is that the subclass writes a header and main parameters,
        # then adds the below string.
        lines = []
        if self.remove_moment is not None:
            lines.append('    Remove moments up to L=%d' % self.remove_moment)
        if self.use_charge_center:
            lines.append('    Compensate for charged system using center of '
                         'majority charge')
        return '\n'.join(lines)

    def solve(self, phi, rho, charge=None, maxcharge=1e-6,
              zero_initial_phi=False, timer=NullTimer()):
        self._init()
        assert np.all(phi.shape == self.gd.n_c)
        assert np.all(rho.shape == self.gd.n_c)

        actual_charge = self.gd.integrate(rho)
        background = (actual_charge / self.gd.dv /
                      self.gd.get_size_of_global_array().prod())

        if self.remove_moment:
            assert not self.gd.pbc_c.any()
            if not hasattr(self, 'gauss'):
                self.gauss = Gaussian(self.gd)
            rho_neutral = rho.copy()
            phi_cor_L = []
            for L in range(self.remove_moment):
                phi_cor_L.append(self.gauss.remove_moment(rho_neutral, L))
            # Remove multipoles for better initial guess
            for phi_cor in phi_cor_L:
                phi -= phi_cor

            niter = self.solve_neutral(phi, rho_neutral, timer=timer)
            # correct error introduced by removing multipoles
            for phi_cor in phi_cor_L:
                phi += phi_cor

            return niter
        if charge is None:
            charge = actual_charge
        if abs(charge) <= maxcharge:
            return self.solve_neutral(phi, rho - background, timer=timer)

        elif abs(charge) > maxcharge and self.gd.pbc_c.all():
            # System is charged and periodic. Subtract a homogeneous
            # background charge

            # Set initial guess for potential
            if zero_initial_phi:
                phi[:] = 0.0

            iters = self.solve_neutral(phi, rho - background, timer=timer)
            return iters

        elif abs(charge) > maxcharge and not self.gd.pbc_c.any():
            # The system is charged and in a non-periodic unit cell.
            # Determine the potential by 1) subtract a gaussian from the
            # density, 2) determine potential from the neutralized density
            # and 3) add the potential from the gaussian density.

            # Load necessary attributes

            # use_charge_center: The monopole will be removed at the
            # center of the majority charge, which prevents artificial
            # dipoles.
            # Due to the shape of the Gaussian and it's Fourier-Transform,
            # the Gaussian representing the charge should stay at least
            # 7 gpts from the borders - see:
            # listserv.fysik.dtu.dk/pipermail/gpaw-developers/2015-July/005806.html
            if self.use_charge_center:
                charge_sign = actual_charge / abs(actual_charge)
                rho_sign = rho * charge_sign
                rho_sign[np.where(rho_sign < 0)] = 0
                absolute_charge = self.gd.integrate(rho_sign)
                center = - (self.gd.calculate_dipole_moment(rho_sign) /
                            absolute_charge)
                border_offset = np.inner(self.gd.h_cv, np.array((7, 7, 7)))
                borders = np.inner(self.gd.h_cv, self.gd.N_c)
                borders -= border_offset
                if np.any(center > borders) or np.any(center < border_offset):
                    raise RuntimeError('Poisson solver: '
                                       'center of charge outside borders '
                                       '- please increase box')
                    center[np.where(center > borders)] = borders
                self.load_gauss(center=center)
            else:
                self.load_gauss()

            # Remove monopole moment
            q = actual_charge / np.sqrt(4 * pi)  # Monopole moment
            rho_neutral = rho - q * self.rho_gauss  # neutralized density

            # Set initial guess for potential
            if zero_initial_phi:
                phi[:] = 0.0
            else:
                axpy(-q, self.phi_gauss, phi)  # phi -= q * self.phi_gauss

            # Determine potential from neutral density using standard solver
            niter = self.solve_neutral(phi, rho_neutral, timer=timer)

            # correct error introduced by removing monopole
            axpy(q, self.phi_gauss, phi)  # phi += q * self.phi_gauss

            return niter
        else:
            # System is charged with mixed boundaryconditions
            if self.metallic_electrodes == 'single':
                self.c = 2
                origin_c = [0, 0, 0]
                origin_c[self.c] = self.gd.N_c[self.c]
                drhot_g, dvHt_g, self.correction = dipole_correction(
                    self.c,
                    self.gd,
                    rho,
                    origin_c=origin_c)
                # self.correction *=-1.
                phi -= dvHt_g
                iters = self.solve_neutral(phi, rho + drhot_g, timer=timer)
                phi += dvHt_g
                phi -= self.correction
                self.correction = 0.0

                return iters

            elif self.metallic_electrodes == 'both':
                iters = self.solve_neutral(phi, rho, timer=timer)
                return iters

            else:
                # System is charged with mixed boundaryconditions
                msg = ('Charged systems with mixed periodic/zero'
                       ' boundary conditions')
                raise NotImplementedError(msg)

    def load_gauss(self, center=None):
        if not hasattr(self, 'rho_gauss') or center is not None:
            gauss = Gaussian(self.gd, center=center)
            self.rho_gauss = gauss.get_gauss(0)
            self.phi_gauss = gauss.get_gauss_pot(0)


class FDPoissonSolver(BasePoissonSolver):
    def __init__(self, nn=3, relax='J', eps=2e-10, maxiter=1000,
                 remove_moment=None, use_charge_center=False,
                 metallic_electrodes=False):
        super(FDPoissonSolver, self).__init__(
            remove_moment=remove_moment,
            use_charge_center=use_charge_center,
            metallic_electrodes=metallic_electrodes)
        self.eps = eps
        self.relax = relax
        self.nn = nn
        self.maxiter = maxiter

        # Relaxation method
        if relax == 'GS':
            # Gauss-Seidel
            self.relax_method = 1
        elif relax == 'J':
            # Jacobi
            self.relax_method = 2
        else:
            raise NotImplementedError('Relaxation method %s' % relax)

        self.description = None
        self._initialized = False

    def todict(self):
        d = super(FDPoissonSolver, self).todict()
        d.update({'name': 'fd', 'nn': self.nn, 'relax': self.relax,
                  'eps': self.eps})
        return d

    def get_stencil(self):
        return self.nn

    def create_laplace(self, gd, scale=1.0, n=1, dtype=float):
        """Instantiate and return a Laplace operator

        Allows subclasses to change the Laplace operator
        """
        return Laplace(gd, scale, n, dtype)

    def set_grid_descriptor(self, gd):
        # Should probably be renamed initialize
        self.gd = gd
        scale = -0.25 / pi

        if self.nn == 'M':
            if not gd.orthogonal:
                raise RuntimeError('Cannot use Mehrstellen stencil with '
                                   'non orthogonal cell.')

            self.operators = [LaplaceA(gd, -scale)]
            self.B = LaplaceB(gd)
        else:
            self.operators = [self.create_laplace(gd, scale, self.nn)]
            self.B = None

        self.interpolators = []
        self.restrictors = []

        level = 0
        self.presmooths = [2]
        self.postsmooths = [1]

        # Weights for the relaxation,
        # only used if 'J' (Jacobi) is chosen as method
        self.weights = [2.0 / 3.0]

        while level < 8:
            try:
                gd2 = gd.coarsen()
            except ValueError:
                break
            self.operators.append(self.create_laplace(gd2, scale, 1))
            self.interpolators.append(Transformer(gd2, gd))
            self.restrictors.append(Transformer(gd, gd2))
            self.presmooths.append(4)
            self.postsmooths.append(4)
            self.weights.append(1.0)
            level += 1
            gd = gd2

        self.levels = level

        if self.operators[-1].gd.N_c.max() > 36:
            # Try to warn exactly once no matter how one uses the solver.
            if gd.comm.parent is None:
                warn = (gd.comm.rank == 0)
            else:
                warn = (gd.comm.parent.rank == 0)

            if warn:
                warntxt = '\n'.join([POISSON_GRID_WARNING, '',
                                     self.get_description()])
            else:
                warntxt = ('Poisson warning from domain rank %d'
                           % self.gd.comm.rank)

            # Warn from all ranks to avoid deadlocks.
            warnings.warn(warntxt, stacklevel=2)

        self._initialized = False
        # The Gaussians depend on the grid as well so we have to 'unload' them
        if hasattr(self, 'rho_gauss'):
            del self.rho_gauss
            del self.phi_gauss

    def get_description(self):
        name = {1: 'Gauss-Seidel', 2: 'Jacobi'}[self.relax_method]
        coarsest_grid = self.operators[-1].gd.N_c
        coarsest_grid_string = ' x '.join([str(N) for N in coarsest_grid])
        assert self.levels + 1 == len(self.operators)
        lines = ['%s solver with %d multi-grid levels'
                 % (name, self.levels + 1),
                 '    Coarsest grid: %s points' % coarsest_grid_string]
        if coarsest_grid.max() > 24:
            # This friendly warning has lower threshold than the big long
            # one that we print when things are really bad.
            lines.extend(['    Warning: Coarse grid has more than 24 points.',
                          '             More multi-grid levels recommended.'])
        lines.extend(['    Stencil: %s' % self.operators[0].description,
                      '    Max iterations: %d' % self.maxiter])
        lines.extend(['    Tolerance: %e' % self.eps])
        lines.append(super(FDPoissonSolver, self).get_description())
        return '\n'.join(lines)

    def _init(self):
        if self._initialized:
            return
        # Should probably be renamed allocate
        gd = self.gd
        self.rhos = [gd.empty()]
        self.phis = [None]
        self.residuals = [gd.empty()]
        for level in range(self.levels):
            gd2 = gd.coarsen()
            self.phis.append(gd2.empty())
            self.rhos.append(gd2.empty())
            self.residuals.append(gd2.empty())
            gd = gd2
        assert len(self.phis) == len(self.rhos)
        level += 1
        assert level == self.levels

        self.step = 0.66666666 / self.operators[0].get_diagonal_element()
        self.presmooths[level] = 8
        self.postsmooths[level] = 8
        self._initialized = True

    def solve_neutral(self, phi, rho, timer=None):
        self._init()
        self.phis[0] = phi
        eps = self.eps

        if self.B is None:
            self.rhos[0][:] = rho
        else:
            self.B.apply(rho, self.rhos[0])

        niter = 1
        maxiter = self.maxiter
        while self.iterate2(self.step) > eps and niter < maxiter:
            niter += 1
        if niter == maxiter:
            msg = 'Poisson solver did not converge in %d iterations!' % maxiter
            raise PoissonConvergenceError(msg)

        # Set the average potential to zero in periodic systems
        if (self.gd.pbc_c).all():
            phi_ave = self.gd.comm.sum(np.sum(phi.ravel()))
            N_c = self.gd.get_size_of_global_array()
            phi_ave /= np.prod(N_c)
            phi -= phi_ave

        return niter

    def iterate2(self, step, level=0):
        """Smooths the solution in every multigrid level"""
        self._init()

        residual = self.residuals[level]

        if level < self.levels:
            self.operators[level].relax(self.relax_method,
                                        self.phis[level],
                                        self.rhos[level],
                                        self.presmooths[level],
                                        self.weights[level])

            self.operators[level].apply(self.phis[level], residual)
            residual -= self.rhos[level]
            self.restrictors[level].apply(residual,
                                          self.rhos[level + 1])
            self.phis[level + 1][:] = 0.0
            self.iterate2(4.0 * step, level + 1)
            self.interpolators[level].apply(self.phis[level + 1], residual)
            self.phis[level] -= residual

        self.operators[level].relax(self.relax_method,
                                    self.phis[level],
                                    self.rhos[level],
                                    self.postsmooths[level],
                                    self.weights[level])
        if level == 0:
            self.operators[level].apply(self.phis[level], residual)
            residual -= self.rhos[level]
            error = self.gd.comm.sum(np.dot(residual.ravel(),
                                            residual.ravel())) * self.gd.dv

            # How about this instead:
            # error = self.gd.comm.max(abs(residual).max())

            return error

    def estimate_memory(self, mem):
        # XXX Memory estimate works only for J and GS, not FFT solver
        # Poisson solver appears to use same amount of memory regardless
        # of whether it's J or GS, which is a bit strange

        gdbytes = self.gd.bytecount()
        nbytes = -gdbytes  # No phi on finest grid, compensate ahead
        for level in range(self.levels):
            nbytes += 3 * gdbytes  # Arrays: rho, phi, residual
            gdbytes //= 8
        mem.subnode('rho, phi, residual [%d levels]' % self.levels, nbytes)

    def __repr__(self):
        template = 'FDPoissonSolver(relax=\'%s\', nn=%s, eps=%e)'
        representation = template % (self.relax, repr(self.nn), self.eps)
        return representation


class NoInteractionPoissonSolver(_PoissonSolver):
    relax_method = 0
    nn = 1

    def get_description(self):
        return 'No interaction'

    def get_stencil(self):
        return 1

    def solve(self, phi, rho, charge, timer=None):
        return 0

    def set_grid_descriptor(self, gd):
        pass

    def todict(self):
        return {'name': 'nointeraction'}

    def estimate_memory(self, mem):
        pass


class FFTPoissonSolver(BasePoissonSolver):
    """FFT Poisson solver for general unit cells."""

    relax_method = 0
    nn = 999

    def __init__(self):
        super(FFTPoissonSolver, self).__init__()
        self._initialized = False

    def get_description(self):
        return 'Parallel FFT'

    def todict(self):
        return {'name': 'fft'}

    def set_grid_descriptor(self, gd):
        # We will probably want to use this on non-periodic grids too...
        assert gd.pbc_c.all()
        self.gd = gd

        self.grids = [gd]
        for c in range(3):
            N_c = gd.N_c.copy()
            N_c[c] = 1  # Will be serial in that direction
            parsize_c = decompose_domain(N_c, gd.comm.size)
            self.grids.append(gd.new_descriptor(parsize_c=parsize_c))

    def _init(self):
        if self._initialized:
            return

        gd = self.grids[-1]
        k2_Q, N3 = construct_reciprocal(gd)
        self.poisson_factor_Q = 4.0 * np.pi / k2_Q
        self._initialized = True

    def solve_neutral(self, phi_g, rho_g, timer=None):
        self._init()
        # Will be a bit more efficient if reduced dimension is always
        # contiguous.  Probably more things can be improved...

        gd1 = self.gd
        work1_g = rho_g

        for c in range(3):
            gd2 = self.grids[c + 1]
            work2_g = gd2.empty(dtype=work1_g.dtype)
            grid2grid(gd1.comm, gd1, gd2, work1_g, work2_g)
            work1_g = fftn(work2_g, axes=[c])
            gd1 = gd2

        work1_g *= self.poisson_factor_Q

        for c in [2, 1, 0]:
            gd2 = self.grids[c]
            work2_g = ifftn(work1_g, axes=[c])
            work1_g = gd2.empty(dtype=work2_g.dtype)
            grid2grid(gd1.comm, gd1, gd2, work2_g, work1_g)
            gd1 = gd2

        phi_g[:] = work1_g.real
        return 1

    def estimate_memory(self, mem):
        mem.subnode('k squared', self.grids[-1].bytecount())


"""def rfst2(A_g, axes=[0,1]):
    assert axes[0] == 0
    assert axes[1] == 1
    x,y,z = A_g.shape
    temp_g = np.zeros((x*2+2, y*2+2, z))
    temp_g[1:x+1, 1:y+1,:] = A_g
    temp_g[x+2:, 1:y+1,:] = -A_g[::-1, :, :]
    temp_g[1:x+1, y+2:,:] = -A_g[:, ::-1, :]
    temp_g[x+2:, y+2:,:] = A_g[::-1, ::-1, :]
    X = -4*rfft2(temp_g, axes=axes)
    return X[1:x+1, 1:y+1, :].real

def irfst2(A_g, axes=[0,1]):
    assert axes[0] == 0
    assert axes[1] == 1
    x,y,z = A_g.shape
    temp_g = np.zeros((x*2+2, (y*2+2)//2+1, z))
    temp_g[1:x+1, 1:y+1,:] = A_g
    temp_g[x+2:, 1:y+1,:] = -A_g[::-1, :, :]
    return -0.25*irfft2(temp_g, axes=axes)[1:x+1, 1:y+1, :].real
"""


use_scipy_transforms = True


def rfst2(A_g, axes=[0, 1]):
    all = set([0, 1, 2])
    third = [all.difference(set(axes)).pop()]

    if use_scipy_transforms:
        Y = A_g
        for axis in axes:
            Y = scipydst(Y, axis=axis, type=1)
        Y *= 2**len(axes)
        return Y

    A_g = np.transpose(A_g, axes + third)
    x, y, z = A_g.shape
    temp_g = np.zeros((x * 2 + 2, y * 2 + 2, z))
    temp_g[1:x + 1, 1:y + 1, :] = A_g
    temp_g[x + 2:, 1:y + 1, :] = -A_g[::-1, :, :]
    temp_g[1:x + 1, y + 2:, :] = -A_g[:, ::-1, :]
    temp_g[x + 2:, y + 2:, :] = A_g[::-1, ::-1, :]
    X = -4 * rfft2(temp_g, axes=[0, 1])[1:x + 1, 1:y + 1, :].real
    return np.transpose(X, np.argsort(axes + third))


def irfst2(A_g, axes=[0, 1]):
    if use_scipy_transforms:
        Y = A_g
        for axis in axes:
            Y = scipydst(Y, axis=axis, type=1)
        magic = 1.0 / (16 * np.prod([A_g.shape[axis] + 1 for axis in axes]))
        Y *= magic
        # Y /= 211200
        return Y

    all = set([0, 1, 2])
    third = [all.difference(set(axes)).pop()]
    A_g = np.transpose(A_g, axes + third)
    x, y, z = A_g.shape
    temp_g = np.zeros((x * 2 + 2, (y * 2 + 2) // 2 + 1, z))
    temp_g[1:x + 1, 1:y + 1, :] = A_g.real
    temp_g[x + 2:, 1:y + 1, :] = -A_g[::-1, :, :].real
    X = -0.25 * irfft2(temp_g, axes=[0, 1])[1:x + 1, 1:y + 1, :]

    T = np.transpose(X, np.argsort(axes + third))
    return T


# This method needs to be taken from fftw / scipy to gain speedup of ~4x
def fst(A_g, axis):
    x, y, z = A_g.shape
    N_c = np.array([x, y, z])
    N_c[axis] = N_c[axis] * 2 + 2
    temp_g = np.zeros(N_c, dtype=A_g.dtype)
    if axis == 0:
        temp_g[1:x + 1, :, :] = A_g
        temp_g[x + 2:, :, :] = -A_g[::-1, :, :]
    elif axis == 1:
        temp_g[:, 1:y + 1, :] = A_g
        temp_g[:, y + 2:, :] = -A_g[:, ::-1, :]
    elif axis == 2:
        temp_g[:, :, 1:z + 1] = A_g
        temp_g[:, :, z + 2:] = -A_g[:, ::, ::-1]
    else:
        raise NotImplementedError()
    X = 0.5j * fft(temp_g, axis=axis)
    if axis == 0:
        return X[1:x + 1, :, :]
    elif axis == 1:
        return X[:, 1:y + 1, :]
    elif axis == 2:
        return X[:, :, 1:z + 1]


def ifst(A_g, axis):
    x, y, z = A_g.shape
    N_c = np.array([x, y, z])
    N_c[axis] = N_c[axis] * 2 + 2
    temp_g = np.zeros(N_c, dtype=A_g.dtype)

    if axis == 0:
        temp_g[1:x + 1, :, :] = A_g
        temp_g[x + 2:, :, :] = -A_g[::-1, :, :]
    elif axis == 1:
        temp_g[:, 1:y + 1, :] = A_g
        temp_g[:, y + 2:, :] = -A_g[:, ::-1, :]
    elif axis == 2:
        temp_g[:, :, 1:z + 1] = A_g
        temp_g[:, :, z + 2:] = -A_g[:, ::, ::-1]
    else:
        raise NotImplementedError()

    X_g = ifft(temp_g, axis=axis)
    if axis == 0:
        return -2j * X_g[1:x + 1, :, :]
    elif axis == 1:
        return -2j * X_g[:, 1:y + 1, :]
    elif axis == 2:
        return -2j * X_g[:, :, 1:z + 1]


def transform(A_g, axis=None, pbc=True):
    if pbc:
        if A_g.size == 0:
            return A_g.astype(complex)

        return fft(A_g, axis=axis)
    else:
        if A_g.size == 0:
            return A_g

        if not use_scipy_transforms:
            x = fst(A_g, axis)
            return x
        y = scipydst(A_g, axis=axis, type=1)
        y *= .5
        return y


def transform2(A_g, axes=None, pbc=[True, True]):
    if all(pbc):
        if A_g.size == 0:
            return A_g.astype(complex)

        return fft2(A_g, axes=axes)
    elif not any(pbc):
        if A_g.size == 0:
            return A_g

        return rfst2(A_g, axes=axes)
    else:
        return transform(transform(A_g, axis=axes[0], pbc=pbc[0]),
                         axis=axes[1], pbc=pbc[1])


def itransform(A_g, axis=None, pbc=True):
    if pbc:
        if A_g.size == 0:
            return A_g.astype(complex)

        return ifft(A_g, axis=axis)
    else:
        if A_g.size == 0:
            return A_g

        if not use_scipy_transforms:
            x = ifst(A_g, axis)
            return x
        y = scipydst(A_g, axis=axis, type=1)
        magic = 1.0 / (A_g.shape[axis] + 1)
        y *= magic
        return y


def itransform2(A_g, axes=None, pbc=[True, True]):
    if all(pbc):
        if A_g.size == 0:
            return A_g.astype(complex)

        return ifft2(A_g, axes=axes)
    elif not any(pbc):
        if A_g.size == 0:
            return A_g

        return irfst2(A_g, axes=axes)
    else:
        return itransform(itransform(A_g, axis=axes[0], pbc=pbc[0]),
                          axis=axes[1], pbc=pbc[1])


class BadAxesError(ValueError):
    pass


class FastPoissonSolver(BasePoissonSolver):
    def __init__(self, nn=3, **kwargs):
        BasePoissonSolver.__init__(self, **kwargs)
        self.nn = nn
        # We may later enable this to work with Cholesky, but not now:
        self.use_cholesky = False

    def _init(self):
        pass

    def set_grid_descriptor(self, gd):
        self.gd = gd
        axes = np.arange(3)
        pbc_c = np.array(gd.pbc_c, dtype=bool)
        periodic_axes = axes[pbc_c]
        non_periodic_axes = axes[np.logical_not(pbc_c)]

        # Find out which axes are orthogonal (0, 1 or 3)
        # Note that one expects that the axes are always rotated in
        # conventional form, thus for all axes to be
        # classified as orthogonal, the cell_cv needs to be diagonal.
        # This may always be achieved by rotating
        # the unit-cell along with the atoms. The classification is
        # inherited from grid_descriptor.orthogonal.
        dotprods = np.dot(gd.cell_cv, gd.cell_cv.T)
        # For each direction, check whether there is only one nonzero
        # element in that row (necessarily being the diagonal element,
        # since this is a cell vector length and must be > 0).
        orthogonal_c = (np.abs(dotprods) > 1e-10).sum(axis=0) == 1
        assert sum(orthogonal_c) in [0, 1, 3]

        non_orthogonal_axes = axes[np.logical_not(orthogonal_c)]

        if not all(pbc_c | orthogonal_c):
            raise BadAxesError('Each axis must be periodic or orthogonal '
                               'to other axes.  But we have pbc={} '
                               'and orthogonal={}'
                               .format(pbc_c.astype(int),
                                       orthogonal_c.astype(int)))

        # We sort them, and pick the longest non-periodic axes as the
        # cholesky axis.
        sorted_non_periodic_axes = sorted(non_periodic_axes,
                                          key=lambda c: gd.N_c[c])
        if self.use_cholesky:
            if len(sorted_non_periodic_axes) > 0:
                cholesky_axes = [sorted_non_periodic_axes[-1]]
                if cholesky_axes[0] in non_orthogonal_axes:
                    msg = ('Cholesky axis cannot be non-orthogonal. '
                           'Do you really want a non-orthogonal non-periodic '
                           'axis? If so, run with use_cholesky=False.')
                    raise NotImplementedError(msg)
                fst_axes = sorted_non_periodic_axes[0:-1]
            else:
                cholesky_axes = []
                fst_axes = []
        else:
            cholesky_axes = []
            fst_axes = sorted_non_periodic_axes
        fft_axes = list(periodic_axes)

        (self.cholesky_axes, self.fst_axes,
         self.fft_axes) = cholesky_axes, fst_axes, fft_axes

        fftfst_axes = self.fft_axes + self.fst_axes
        axes = self.fft_axes + self.fst_axes + self.cholesky_axes
        self.axes = axes

        # Create xy flat decomposition (where x=axes[0] and y=axes[1])
        parsize_c = [1, 1, 1]
        parsize_c[axes[2]] = gd.comm.size
        gd1d = gd.new_descriptor(parsize_c=parsize_c,
                                 allow_empty_domains=True)
        self.gd1d = gd1d

        # Create z flat decomposition
        domain = gd.N_c.copy()
        domain[axes[2]] = 1
        parsize_c = decompose_domain(domain, gd.comm.size)
        gd2d = gd.new_descriptor(parsize_c=parsize_c)
        self.gd2d = gd2d

        # Calculate eigenvalues in fst/fft decomposition for
        # non-cholesky axes in parallel
        r_cx = np.indices(gd2d.n_c)
        r_cx += gd2d.beg_c[:, np.newaxis, np.newaxis, np.newaxis]
        r_cx = r_cx.astype(complex)
        for c, axis in enumerate(fftfst_axes):
            r_cx[axis] *= 2j * np.pi / gd2d.N_c[axis]
            if axis in fst_axes:
                r_cx[axis] /= 2
        for c, axis in enumerate(cholesky_axes):
            r_cx[axis] = 0.0
        np.exp(r_cx, out=r_cx)
        fft_lambdas = np.zeros_like(r_cx[0], dtype=complex)
        laplace = Laplace(self.gd, -0.25 / pi, self.nn)
        self.stencil_description = laplace.description

        for coeff, offset_c in zip(laplace.coef_p, laplace.offset_pc):
            offset_c = np.array(offset_c)
            if not any(offset_c):
                # The centerpoint is handled with (temp-1.0)
                continue
            non_zero_axes, = np.where(offset_c)
            if set(non_zero_axes).issubset(fftfst_axes):
                temp = np.ones_like(fft_lambdas)
                for c, axis in enumerate(fftfst_axes):
                    temp *= r_cx[axis] ** offset_c[axis]
                fft_lambdas += coeff * (temp - 1.0)

        assert np.linalg.norm(fft_lambdas.imag) < 1e-10
        fft_lambdas = fft_lambdas.real.copy()  # arr.real is not contiguous

        # If there is no Cholesky decomposition, the system is already
        # fully diagonal and we can directly invert the linear problem
        # by dividing with the eigenvalues.
        assert len(cholesky_axes) == 0
        # if len(cholesky_axes) == 0:
        with np.errstate(divide='ignore'):
            self.inv_fft_lambdas = np.where(
                np.abs(fft_lambdas) > 1e-10, 1.0 / fft_lambdas, 0)

    def solve_neutral(self, phi_g, rho_g, timer=None):
        if len(self.cholesky_axes) != 0:
            raise NotImplementedError

        gd = self.gd
        gd1d = self.gd1d
        gd2d = self.gd2d
        comm = self.gd.comm
        axes = self.axes

        with timer('Communicate to 1D'):
            work1d_g = gd1d.empty(dtype=rho_g.dtype)
            grid2grid(comm, gd, gd1d, rho_g, work1d_g)
        with timer('FFT 2D'):
            work1d_g = transform2(work1d_g, axes=axes[:2],
                                  pbc=gd.pbc_c[axes[:2]])
        with timer('Communicate to 2D'):
            work2d_g = gd2d.empty(dtype=work1d_g.dtype)
            grid2grid(comm, gd1d, gd2d, work1d_g, work2d_g)
        with timer('FFT 1D'):
            work2d_g = transform(work2d_g, axis=axes[2],
                                 pbc=gd.pbc_c[axes[2]])

        # The remaining problem is 0D dimensional, i.e the problem
        # has been fully diagonalized
        work2d_g *= self.inv_fft_lambdas

        with timer('FFT 1D'):
            work2d_g = itransform(work2d_g, axis=axes[2],
                                  pbc=gd.pbc_c[axes[2]])
        with timer('Communicate from 2D'):
            work1d_g = gd1d.empty(dtype=work2d_g.dtype)
            grid2grid(comm, gd2d, gd1d, work2d_g, work1d_g)
        with timer('FFT 2D'):
            work1d_g = itransform2(work1d_g, axes=axes[1::-1],
                                   pbc=gd.pbc_c[axes[1::-1]])
        with timer('Communicate from 1D'):
            work_g = gd.empty(dtype=work1d_g.dtype)
            grid2grid(comm, gd1d, gd, work1d_g, work_g)

        phi_g[:] = work_g.real
        return 1  # Non-iterative method, return 1 iteration

    def todict(self):
        d = super(FastPoissonSolver, self).todict()
        d.update({'name': 'fast', 'nn': self.nn})
        return d

    def estimate_memory(self, mem):
        pass

    def get_description(self):
        lines = [f'{self.__class__.__name__} using',
                 f'    Stencil: {self.stencil_description}',
                 f'    FFT axes: {self.fft_axes}',
                 f'    FST axes: {self.fst_axes}',
                 ]
        lines.append(BasePoissonSolver.get_description(self))
        return '\n'.join(lines)



from math import exp, pi, sqrt

import numpy as np

from ase.units import Bohr, Hartree, alpha
from gpaw.pes import ds_prefactor


class State:
    """Electronic state base class."""

    def __init__(self, gd):
        self.gd = gd
        self.r_c = []
        for c in range(3):
            r_G = (np.arange(gd.n_c[c], dtype=float) +
                   gd.beg_c[c]) * gd.h_cv.diagonal()[c]
            self.r_c.append(r_G)

    def get_grid(self):
        return self.grid_g

    def set_grid(self, grid_g):
        #        assert(grid_g.shape == self.gd.empty().shape)
        self.grid_g = grid_g

    def get_energy(self):
        """The states energy in [eV]"""
        return self.energy * Hartree

    def set_energy(self, energy):
        """Set the states energy in [eV]"""
        self.energy = energy / Hartree


class BoundState(State):

    """Bound state originating from a gpaw calculation."""

    def __init__(self,
                 calculator,
                 s, n):
        self.gd = calculator.wfs.gd
        self.kpt = calculator.wfs.kpt_u[s]
        self.grid_g = self.kpt.psit_nG[n]
        self.energy = self.kpt.eps_n[n]


class H1s(State):

    """Analytic H1s state."""

    def __init__(self, gd=None, center=None, Z=1):
        if center is None:
            center = 0.5 * gd.n_c * gd.h_c
        else:
            self.center = center / Bohr
        self.Z = Z
        self.energy = - Z ** 2 / 2.
        if gd is not None:
            self.grid_g = None
            State.__init__(self, gd)

    def get_grid(self):
        if self.grid_g is None:
            gd = self.gd
            assert gd.orthogonal
            wf = gd.zeros(dtype=float)
            vr0 = np.array([0., 0., 0.])
            for i in range(gd.n_c[0]):
                vr0[0] = (gd.beg_c[0] + i) * gd.h_cv[0, 0]
                for j in range(gd.n_c[1]):
                    vr0[1] = (gd.beg_c[1] + j) * gd.h_cv[1, 1]
                    for k in range(gd.n_c[2]):
                        vr0[2] = (gd.beg_c[2] + k) * gd.h_cv[2, 2]
                        vr = vr0 - self.center
                        r = sqrt(np.dot(vr, vr))
                        wf[i, j, k] = exp(-self.Z * r)
            self.grid_g = wf * sqrt(self.Z ** 3 / pi)
        return self.grid_g

    def get_me_c(self, k_c, form='L'):
        """Transition matrix element."""
        k = sqrt(np.dot(k_c, k_c))
        if form == 'L':
            pre = 2j
        elif form == 'V':
            pre = 1j
        else:
            raise NotImplementedError
        return pre * k_c * self.FT(k)

    def get_ds(self, Ekin, form='L', units='Mb'):
        """Angular averaged cross section.

        Ekin: photoelectron kinetic energy [eV]"""
        E = Ekin / Hartree
        k = sqrt(2 * E)
        omega = E - self.energy
        k_c = np.array([0., 0., k])
        me_c = self.get_me_c(k_c, form)
        T2 = np.abs(np.dot(me_c, me_c)) / 3.
        pre = ds_prefactor[units]
        return pre * (2 * pi) ** 2 * alpha * k * 4 * pi / omega * T2

    def FT(self, k):
        return sqrt(8 * self.Z ** 5) / pi / (k ** 2 + self.Z ** 2) ** 2



from ase import Atoms
from gpaw import GPAW
from gpaw.mixer import FFTMixer
from gpaw import PW
from gpaw.test import equal


def test_pw_fftmixer():
    bulk = Atoms('Li', pbc=True,
                 cell=[2.6, 2.6, 2.6])
    k = 4
    bulk.calc = GPAW(mode=PW(200),
                     kpts=(k, k, k),
                     mixer=FFTMixer(),
                     eigensolver='rmm-diis')
    e = bulk.get_potential_energy()
    equal(e, -1.98481281259, 1.0e-4)
